from typing import Tuple, Set, Union, MutableMapping, Any, Mapping, Iterable, Generator, List

from datapipelines import Query, PipelineContext, QueryValidationError

from ..data import Region, Platform, Queue, Tier, Division, Position

from ..dto.champion import ChampionRotationDto
from ..dto.championmastery import ChampionMasteryDto, ChampionMasteryListDto, ChampionMasteryScoreDto
from ..dto.league import LeagueEntriesDto, LeagueSummonerEntriesDto, LeagueEntryDto
from ..dto.staticdata import ChampionDto, ChampionListDto, ItemDto, ItemListDto, LanguageStringsDto, LanguagesDto, ProfileIconDataDto, ProfileIconDetailsDto, RealmDto, RuneDto, RuneListDto, SummonerSpellDto, SummonerSpellListDto, MapDto, MapListDto, VersionListDto
from ..dto.status import ShardStatusDto
from ..dto.match import MatchDto, MatchReferenceDto, TimelineDto
from ..dto.spectator import CurrentGameInfoDto, FeaturedGamesDto
from ..dto.summoner import SummonerDto

from ..core.championmastery import ChampionMastery, ChampionMasteries
from ..core.league import LeagueSummonerEntries, ChallengerLeague, GrandmasterLeague, MasterLeague, League, LeagueEntry
from ..core.staticdata import Champion, Rune, Item, SummonerSpell, Map, Locales, LanguageStrings, ProfileIcon, ProfileIcons, Realms, Versions, Items, Champions, Maps, SummonerSpells, Runes
from ..core.status import ShardStatus
from ..core.match import Match, MatchHistory, Timeline
from ..core.summoner import Summoner
from ..core.spectator import CurrentMatch, FeaturedMatches, CurrentGameParticipantData
from ..core.champion import ChampionRotation, ChampionRotationData

from ..core.staticdata.champion import ChampionData
from ..core.staticdata.item import ItemData
from ..core.staticdata.summonerspell import SummonerSpellData
from ..core.staticdata.rune import RuneData
from ..core.staticdata.map import MapData
from ..core.summoner import SummonerData

#############
# Utilities #
#############


def _rgetattr(obj, key):
    """Recursive getattr for handling dots in keys."""
    for k in key.split("."):
        obj = getattr(obj, k)
    return obj

def _hash_included_data(included_data: Set[str]) -> int:
    return hash(tuple(included_data))


def _get_default_version(query: Mapping[str, Any], context: PipelineContext) -> str:
    try:
        pipeline = context[PipelineContext.Keys.PIPELINE]
        versions = pipeline.get(VersionListDto, {"platform": query["platform"]})
        return versions["versions"][0]
    except TypeError as error:
        raise KeyError("`version` must be provided in query") from error


def _get_default_locale(query: Mapping[str, Any], context: PipelineContext) -> str:
    return query["platform"].default_locale


def _region_to_platform_generator(regions: Iterable[Region]) -> Generator[Platform, None, None]:
    for region in regions:
        try:
            yield Region(region).platform
        except ValueError as e:
            raise QueryValidationError from e


def convert_region_to_platform(query: MutableMapping[str, Any]) -> None:
    if "region" in query and "platform" not in query:
        try:
            query["platform"] = Region(query["region"]).platform
        except ValueError as e:
            raise QueryValidationError from e

    if "regions" in query and "platforms" not in query:
        query["platforms"] = _region_to_platform_generator(query["regions"])

    if "region" in query and not isinstance(query["region"], Region):
        query["region"] = Region(query["region"])


#######
# DTO #
#######


################
# Champion API #
################


validate_champion_rotation_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_champion_rotation_dto_query = Query. \
    has("platform").as_(Platform)


def for_champion_rotation_dto(champion_rotation: ChampionRotationDto) -> str:
    return champion_rotation["platform"]


def for_champion_rotation_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_champion_rotation_dto_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


########################
# Champion Mastery API #
########################


validate_champion_mastery_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerId").as_(str).also. \
    has("championId").as_(int)


validate_many_champion_mastery_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerId").as_(str).also. \
    has("championIds").as_(Iterable)


def for_champion_mastery_dto(champion_mastery: ChampionMasteryDto) -> Tuple[str, str, int]:
    return champion_mastery["platform"], champion_mastery["playerId"], champion_mastery["championId"]


def for_champion_mastery_dto_query(query: Query) -> Tuple[str, str, int]:
    return query["platform"].value, query["playerId"], query["championId"]


def for_many_champion_mastery_dto_query(query: Query) -> Generator[Tuple[str, str, int], None, None]:
    for champion_id in query["championIds"]:
        try:
            champion_id = int(champion_id)
            yield query["platform"].value, query["playerId"], champion_id
        except ValueError as e:
            raise QueryValidationError from e


validate_champion_mastery_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerId").as_(str)


validate_many_champion_mastery_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerIds").as_(Iterable)


def for_champion_mastery_list_dto(champion_mastery_list: ChampionMasteryListDto) -> Tuple[str, str]:
    return champion_mastery_list["platform"], champion_mastery_list["playerId"]


def for_champion_mastery_list_dto_query(query: Query) -> Tuple[str, str]:
    return query["platform"].value, query["playerId"]


def for_many_champion_mastery_list_dto_query(query: Query) -> Generator[Tuple[str, str], None, None]:
    for summoner_id in query["playerIds"]:
        try:
            yield query["platform"].value, summoner_id
        except ValueError as e:
            raise QueryValidationError from e


validate_champion_mastery_score_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerId").as_(str)


validate_many_champion_mastery_score_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("playerIds").as_(Iterable)


def for_champion_mastery_score_dto(champion_mastery_score: ChampionMasteryScoreDto) -> Tuple[str, str]:
    return champion_mastery_score["platform"], champion_mastery_score["playerId"]


def for_champion_mastery_score_dto_query(query: Query) -> Tuple[str, str]:
    return query["platform"].value, query["playerId"]


def for_many_champion_mastery_score_dto_query(query: Query) -> Generator[Tuple[str, str], None, None]:
    for summoner_id in query["playerIds"]:
        try:
            yield query["platform"].value, summoner_id
        except ValueError as e:
            raise QueryValidationError from e


##############
# League API #
##############

validate_league_entries_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("queue").as_(Queue).also. \
    has("tier").as_(Tier).also. \
    has("page").as_(int).also. \
    has("id").as_(int)  # League ID


def for_league_entries_dto(league_entries: LeagueEntriesDto) -> Tuple[str, str, str, int, int]:
    return league_entries["platform"], league_entries["queue"], league_entries["tier"], league_entries["id"], league_entries["page"]


def for_league_entries_dto_query(query: Query) -> Tuple[str, str, str, int, int]:
    return query["platform"].value, query["queue"].value, query["tier"].value, query["id"], query["page"]


validate_league_summoner_entries_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("id").as_(int)  # Summoner ID

def for_league_summoner_entries_dto(league_entries: LeagueEntriesDto) -> Tuple[str, int]:
    return league_entries["platform"], league_entries["id"]


def for_league_summoner_entries_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["id"]


###################
# Static Data API #
###################


# Champion

validate_champion_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_champion_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_champion_dto(champion: ChampionDto, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return champion["platform"], champion["version"], champion["locale"], _hash_included_data(champion["includedData"]), champion[identifier]


def for_champion_dto_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"]), query[identifier]


def for_many_champion_dto_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["includedData"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_champion_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_champion_list_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_champion_list_dto(champion_list: ChampionListDto) -> Tuple[str, str, str, int]:
    return champion_list["platform"], champion_list["version"], champion_list["locale"], _hash_included_data(champion_list["includedData"])


def for_champion_list_dto_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"])


def for_many_champion_list_dto_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Item


validate_item_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_item_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("name").as_(Iterable)


def for_item_dto(item: ItemDto, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return item["platform"], item["version"], item["locale"], _hash_included_data(item["includedData"]), item[identifier]


def for_item_dto_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"]), query[identifier]


def for_many_item_dto_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["includedData"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_item_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_item_list_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_item_list_dto(item_list: ItemListDto) -> Tuple[str, str, str, int]:
    return item_list["platform"], item_list["version"], item_list["locale"], _hash_included_data(item_list["includedData"])


def for_item_list_dto_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"])


def for_many_item_list_dto_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Language


validate_language_strings_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


validate_many_language_strings_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


def for_language_strings_dto(language_strings: LanguageStringsDto) -> Tuple[str, str, str]:
    return language_strings["platform"], language_strings["version"], language_strings["locale"]


def for_language_strings_dto_query(query: Query) -> Tuple[str, str, str]:
    return query["platform"].value, query["version"], query["locale"]


def for_many_language_strings_dto_query(query: Query) -> Generator[Tuple[str, str, str], None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"]
        except ValueError as e:
            raise QueryValidationError from e


validate_languages_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_languages_dto_query = Query. \
    has("platforms").as_(Iterable)


def for_languages_dto(languages: LanguagesDto) -> str:
    return languages["platform"]


def for_languages_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_languages_dto_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


# Map

validate_map_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("mapId").as_(int).or_("mapName").as_(str)


validate_many_map_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("mapIds").as_(Iterable).or_("mapNames").as_(Iterable)


def for_map_dto(map: MapDto, identifier: str = "mapId") -> Tuple[str, str, str, Union[int, str]]:
    return map["platform"], map["version"], map["locale"], map[identifier]


def for_map_dto_query(query: Query) -> Tuple[str, str, str, Union[int, str]]:
    identifier = "mapId" if "mapId" in query else "mapName"
    return query["platform"].value, query["version"], query["locale"], query[identifier]


def for_many_map_dto_query(query: Query) -> Generator[Tuple[str, str, str, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["mapIds"], int) if "mapIds" in query else (query["mapNames"], str)
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_map_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


validate_many_map_list_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


def for_map_list_dto(map_list: MapListDto) -> Tuple[str, str, str]:
    return map_list["platform"], map_list["version"], map_list["locale"]


def for_map_list_dto_query(query: Query) -> Tuple[str, str, str]:
    return query["platform"].value, query["version"], query["locale"]


def for_many_map_list_dto_query(query: Query) -> Generator[Tuple[str, str, str], None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"]
        except ValueError as e:
            raise QueryValidationError from e


# Profile Icon


validate_profile_icon_data_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


validate_many_profile_icon_data_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


def for_profile_icon_data_dto(profile_icon_data: ProfileIconDataDto) -> Tuple[str, str, str]:
    return profile_icon_data["platform"], profile_icon_data["version"], profile_icon_data["locale"]


def for_profile_icon_data_dto_query(query: Query) -> Tuple[str, str, str]:
    return query["platform"].value, query["version"], query["locale"]


def for_many_profile_icon_data_dto_query(query: Query) -> Generator[Tuple[str, str, str], None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"]
        except ValueError as e:
            raise QueryValidationError from e


validate_profile_icon_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("id").as_(int)


validate_many_profile_icon_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("ids").as_(Iterable)


def for_profile_icon_dto(profile_icon: ProfileIconDetailsDto) -> Tuple[str, str, str, int]:
    return profile_icon["platform"], profile_icon["version"], profile_icon["locale"], profile_icon["id"]


def for_profile_icon_dto_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"].value, query["version"], query["locale"], query["id"]


def for_many_profile_icon_dto_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    for id in query["ids"]:
        try:
            id = int(id)
            yield query["platform"].value, query["version"], query["locale"], id
        except ValueError as e:
            raise QueryValidationError from e


# Realm


validate_realm_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_realm_dto_query = Query. \
    has("platforms").as_(Iterable)


def for_realm_dto(realm: RealmDto) -> str:
    return realm["platform"]


def for_realm_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_realm_dto_query(query: Query) -> Generator[str, None, None]:
    return query["platform"].value


# Rune


validate_rune_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_rune_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_rune_dto(rune: RuneDto, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return rune["platform"], rune["version"], rune["locale"], _hash_included_data(rune["includedData"]), rune[identifier]


def for_rune_dto_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"]), query[identifier]


def for_many_rune_dto_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["includedData"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_rune_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_rune_list_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_rune_list_dto(rune_list: RuneListDto) -> Tuple[str, str, str, int]:
    return rune_list["platform"], rune_list["version"], rune_list["locale"], _hash_included_data(rune_list["includedData"])


def for_rune_list_dto_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"])


def for_many_rune_list_dto_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Summoner Spell


validate_summoner_spell_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_summoner_spell_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_summoner_spell_dto(summoner_spell: SummonerSpellDto, identifier: str = "id") -> Tuple[str, str, str, int, Union[int, str]]:
    return summoner_spell["platform"], summoner_spell["version"], summoner_spell["locale"], _hash_included_data(summoner_spell["includedData"]), summoner_spell[identifier]


def for_summoner_spell_dto_query(query: Query) -> Tuple[str, str, str, int, Union[int, str]]:
    identifier = "id" if "id" in query else "name"
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"]), query[identifier]


def for_many_summoner_spell_dto_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    identifiers, identifier_type = (query["ids"], int) if "ids" in query else (query["names"], str)
    included_data_hash = _hash_included_data(query["includedData"])
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, query["version"], query["locale"], included_data_hash, identifier
        except ValueError as e:
            raise QueryValidationError from e


validate_summoner_spell_list_dto_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_summoner_spell_list_dto_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_summoner_spell_list_dto(summoner_spell_list: SummonerSpellListDto) -> Tuple[str, str, str, int]:
    return summoner_spell_list["platform"], summoner_spell_list["version"], summoner_spell_list["locale"], _hash_included_data(summoner_spell_list["includedData"])


def for_summoner_spell_list_dto_query(query: Query) -> Tuple[str, str, str, int]:
    return query["platform"].value, query["version"], query["locale"], _hash_included_data(query["includedData"])


def for_many_summoner_spell_list_dto_query(query: Query) -> Generator[Tuple[str, str, str, int], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value, query["version"], query["locale"], included_data_hash
        except ValueError as e:
            raise QueryValidationError from e


# Version


validate_version_list_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_version_list_dto_query = Query. \
    has("platforms").as_(Iterable)


def for_version_list_dto(version_list: VersionListDto) -> str:
    return version_list["platform"]


def for_version_list_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_version_list_dto_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


##############
# Status API #
##############


validate_shard_status_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_shard_status_dto_query = Query. \
    has("platforms").as_(Iterable)


def for_shard_status_dto(shard_status: ShardStatusDto) -> str:
    return shard_status["platform"]


def for_shard_status_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_shard_status_dto_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


#############
# Match API #
#############


validate_match_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameId").as_(int)


validate_many_match_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameIds").as_(Iterable)


def for_match_dto(match: MatchDto) -> Tuple[str, int]:
    return match["platform"], match["gameId"]


def for_match_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["gameId"]


def for_many_match_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for game_id in query["gameIds"]:
        try:
            game_id = int(game_id)
            yield query["platform"].value, game_id
        except ValueError as e:
            raise QueryValidationError from e


validate_match_reference_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameId").as_(int)


validate_many_match_reference_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameIds").as_(Iterable)


def for_match_reference_dto(match_reference: MatchReferenceDto) -> Tuple[str, int]:
    return match_reference["platform"], match_reference["gameId"]


def for_match_reference_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["gameId"]


def for_many_match_reference_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for game_id in query["gameIds"]:
        try:
            game_id = int(game_id)
            yield query["platform"].value, game_id
        except ValueError as e:
            raise QueryValidationError from e


validate_match_timeline_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("matchId").as_(int)


validate_many_match_timeline_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("matchIds").as_(Iterable)


def for_match_timeline_dto(match_timeline: TimelineDto) -> Tuple[str, int]:
    return match_timeline["platform"], match_timeline["matchId"]


def for_match_timeline_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["matchId"]


def for_many_match_timeline_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for match_id in query["matchIds"]:
        try:
            match_id = int(match_id)
            yield query["platform"].value, match_id
        except ValueError as e:
            raise QueryValidationError from e


#################
# Spectator API #
#################


validate_current_game_info_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameId").as_(int)


validate_many_current_game_info_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("gameIds").as_(Iterable)


def for_current_game_info_dto(current_game_info: CurrentGameInfoDto) -> Tuple[str, int]:
    return current_game_info["platform"], current_game_info["gameId"]


def for_current_game_info_dto_query(query: Query) -> Tuple[str, int]:
    return query["platform"].value, query["gameId"]


def for_many_current_game_info_dto_query(query: Query) -> Generator[Tuple[str, int], None, None]:
    for game_id in query["gameIds"]:
        try:
            game_id = int(game_id)
            yield query["platform"].value, game_id
        except ValueError as e:
            raise QueryValidationError from e


validate_featured_game_dto_query = Query. \
    has("platform").as_(Platform)


validate_many_featured_game_dto_query = Query. \
    has("platforms").as_(Iterable)


def for_featured_games_dto(featured_games: FeaturedGamesDto) -> str:
    return featured_games["platform"]


def for_featured_games_dto_query(query: Query) -> str:
    return query["platform"].value


def for_many_featured_games_dto_query(query: Query) -> Generator[str, None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield platform.value
        except ValueError as e:
            raise QueryValidationError from e


################
# Summoner API #
################


validate_summoner_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("id").as_(int).or_("accountId").as_(int).or_("name").as_(str)


validate_many_summoner_dto_query = Query. \
    has("platform").as_(Platform).also. \
    has("ids").as_(Iterable).or_("accountIds").as_(Iterable).or_("names").as_(Iterable)


def for_summoner_dto(summoner: SummonerDto, identifier: str = "id") -> Tuple[str, Union[int, str]]:
    return summoner["platform"], summoner[identifier]


def for_summoner_dto_query(query: Query) -> Tuple[str, Union[int, str]]:
    if "id" in query:
        identifier = "id"
    elif "accountId" in query:
        identifier = "accountId"
    else:
        identifier = "name"
    return query["platform"].value, query[identifier]


def for_many_summoner_dto_query(query: Query) -> Generator[Tuple[str, Union[int, str]], None, None]:
    if "ids" in query:
        identifiers, identifier_type = query["ids"], int
    elif "accountIds" in query:
        identifiers, identifier_type = query["accountIds"], int
    else:
        identifiers, identifier_type = query["names"], str
    for identifier in identifiers:
        try:
            identifier = identifier_type(identifier)
            yield query["platform"].value, identifier
        except ValueError as e:
            raise QueryValidationError from e


########
# Core #
########


################
# Champion API #
################


validate_champion_rotation_query = Query. \
    has("platform").as_(Platform)


validate_many_champion_rotation_query = Query. \
    has("platform").as_(Platform)


def for_champion_rotation(champion_rotation: ChampionRotationData) -> List[Region]:
    keys = [champion_rotation.platform]
    return keys


def for_champion_rotation_query(query: Query) -> List[str]:
    keys = [query["platform"].value]
    return keys


########################
# Champion Mastery API #
########################


validate_champion_mastery_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.id").as_(str).or_("summoner.accountId").as_(str).or_("summoner.name").as_(str).also. \
    has("champion.id").as_(int).or_("champion.name").as_(str)


validate_many_champion_mastery_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.id").as_(str).or_("summoner.accountId").as_(str).or_("summoner.name").as_(str).also. \
    has("champions.id").as_(Iterable).or_("champions.name").as_(Iterable)


def for_champion_mastery(champion_mastery: ChampionMastery) -> List[Tuple]:
    keys = []
    try:
        keys.append((champion_mastery.platform.value, champion_mastery.summoner._data[SummonerData].id, champion_mastery.champion._data[ChampionData].id))
    except AttributeError:
        pass
    try:
        keys.append((champion_mastery.platform.value, champion_mastery.summoner._data[SummonerData].id, champion_mastery.champion._data[ChampionData].name))
    except AttributeError:
        pass
    try:
        keys.append((champion_mastery.platform.value, champion_mastery.summoner._data[SummonerData].name, champion_mastery.champion._data[ChampionData].id))
    except AttributeError:
        pass
    try:
        keys.append((champion_mastery.platform.value, champion_mastery.summoner._data[SummonerData].name, champion_mastery.champion._data[ChampionData].name))
    except AttributeError:
        pass
    try:
        keys.append((champion_mastery.platform.value, champion_mastery.summoner._data[SummonerData].account_id, champion_mastery.champion._data[ChampionData].id))
    except AttributeError:
        pass
    try:
        keys.append((champion_mastery.platform.value, champion_mastery.summoner._data[SummonerData].account_id, champion_mastery.champion._data[ChampionData].name))
    except AttributeError:
        pass
    return keys


def for_champion_mastery_query(query: Query) -> List[Tuple]:
    keys = []
    if "summoner.id" in query and "champion.id" in query:
        keys.append((query["platform"].value, query["summoner.id"], query["champion.id"]))
    if "summoner.id" in query and "champion.name" in query:
        keys.append((query["platform"].value, query["summoner.id"], query["champion.name"]))
    if "summoner.name" in query and "champion.id" in query:
        keys.append((query["platform"].value, query["summoner.name"], query["champion.id"]))
    if "summoner.name" in query and "champion.name" in query:
        keys.append((query["platform"].value, query["summoner.name"], query["champion.name"]))
    if "summoner.accountId" in query and "champion.id" in query:
        keys.append((query["platform"].value, query["summoner.accountId"], query["champion.id"]))
    if "summoner.accountId" in query and "champion.name" in query:
        keys.append((query["platform"].value, query["summoner.accountId"], query["champion.name"]))
    return keys


def for_many_champion_mastery_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    grouped_identifiers = []
    identifier_types = []
    if "champions.id" in query:
        grouped_identifiers.append(query["ids"])
        identifier_types.append(int)
    if "champions.name" in query:
        grouped_identifiers.append(query["names"])
        identifier_types.append(str)
    for identifiers in zip(*grouped_identifiers):
        keys = []
        for identifier, identifier_type in zip(identifiers, identifier_types):
            identifier = identifier_type(identifier)
            if "summoner.id" in query:
                keys.append((query["platform"].value, query["summoner.id"], identifier))
            if "summoner.name" in query:
                keys.append((query["platform"].value, query["summoner.name"], identifier))
            if "summoner.accountId" in query:
                keys.append((query["platform"].value, query["summoner.accountId"], identifier))
        if len(keys) == 0:
            raise QueryValidationError
        yield keys


validate_champion_masteries_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.id").as_(str).or_("summoner.accountId").as_(int).or_("summoner.name")


validate_many_champion_masteries_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.id").as_(str).or_("summoner.accountId").as_(int).or_("summoner.name")


def for_champion_masteries(champion_mastery: ChampionMasteries) -> List[Tuple]:
    keys = []
    try:
        keys.append((champion_mastery.platform.value, champion_mastery.summoner._data[SummonerData].id))
    except AttributeError:
        pass
    try:
        keys.append((champion_mastery.platform.value, champion_mastery.summoner._data[SummonerData].name))
    except AttributeError:
        pass
    try:
        keys.append((champion_mastery.platform.value, champion_mastery.summoner._data[SummonerData].account_id))
    except AttributeError:
        pass
    return keys


def for_champion_masteries_query(query: Query) -> List[Tuple]:
    keys = []
    if "summoner.id" in query:
        keys.append((query["platform"].value, query["summoner.id"]))
    if "summoner.name" in query:
        keys.append((query["platform"].value, query["summoner.name"]))
    if "summoner.accountId" in query:
        keys.append((query["platform"].value, query["summoner.accountId"]))
    return keys


##############
# League API #
##############


# League Entries

validate_league_entries_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.id").as_(str)


validate_many_league_entries_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoners.id").as_(Iterable)


def for_league_summoner_entries(entries: LeagueSummonerEntries) -> List[Tuple[str, str]]:
    return [(entries.platform.value, entries._LeagueSummonerEntries__summoner.id)]


def for_league_summoner_entries_query(query: Query) -> List[Tuple[str, str]]:
    return [(query["platform"].value, query["summoner.id"])]


def for_many_league_summoner_entries_query(query: Query) -> Generator[List[Tuple[str, str]], None, None]:
    for id in query["summoners.id"]:
        try:
            yield [(query["platform"].value, id)]
        except ValueError as e:
            raise QueryValidationError from e

# Leagues

validate_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("id").as_(str)


validate_many_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("ids").as_(Iterable)


def for_league(league: League) -> List[Tuple[str, str]]:
    return [(league.platform.value, league.id)]


def for_league_query(query: Query) -> List[Tuple[str, str]]:
    return [(query["platform"].value, query["id"])]


def for_many_league_query(query: Query) -> Generator[List[Tuple[str, str]], None, None]:
    for id in query["ids"]:
        try:
            yield [(query["platform"].value, id)]
        except ValueError as e:
            raise QueryValidationError from e

# Challenger

validate_challenger_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("queue").as_(Queue)


validate_many_challenger_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("queues").as_(Iterable)


def for_challenger_league(league: ChallengerLeague) -> List[Tuple[str, str]]:
    return [(league.platform.value, league.queue.value)]


def for_challenger_league_query(query: Query) -> List[Tuple[str, str]]:
    return [(query["platform"].value, query["queue"].value)]


def for_many_challenger_league_query(query: Query) -> Generator[List[Tuple[str, str]], None, None]:
    for queue in query["queues"]:
        try:
            yield [(query["platform"].value, queue.value)]
        except ValueError as e:
            raise QueryValidationError from e

# Grandmaster

validate_grandmaster_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("queue").as_(Queue)


validate_many_grandmaster_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("queues").as_(Iterable)


def for_grandmaster_league(league: GrandmasterLeague) -> List[Tuple[str, str]]:
    return [(league.platform.value, league.queue.value)]


def for_grandmaster_league_query(query: Query) -> List[Tuple[str, str]]:
    return [(query["platform"].value, query["queue"].value)]


def for_many_grandmaster_league_query(query: Query) -> Generator[List[Tuple[str, str]], None, None]:
    for queue in query["queues"]:
        try:
            yield [(query["platform"].value, queue.value)]
        except ValueError as e:
            raise QueryValidationError from e

# Master

validate_master_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("queue").as_(Queue)


validate_many_master_league_query = Query. \
    has("platform").as_(Platform).also. \
    has("queues").as_(Iterable)


def for_master_league(league: MasterLeague) -> List[Tuple[str, str]]:
    return [(league.platform.value, league.queue.value)]


def for_master_league_query(query: Query) -> List[Tuple[str, str]]:
    return [(query["platform"].value, query["queue"].value)]


def for_many_master_league_query(query: Query) -> Generator[List[Tuple[str, str]], None, None]:
    for queue in query["queues"]:
        try:
            yield [(query["platform"].value, queue.value)]
        except ValueError as e:
            raise QueryValidationError from e

# League Entries List

validate_league_entries_list_query = Query. \
    has("queue").as_(Queue).also. \
    has("tier").as_(Tier).also. \
    has("division").as_(Division).also. \
    has("platform").as_(Platform)

def for_league_entries_list(lel: LeagueSummonerEntries) -> List[Tuple[str, str, str, str]]:
    return [(lel.platform.value, lel.queue.value, lel.tier.value, lel.division.value)]

def for_league_entries_list_query(query: Query) -> List[Tuple[str, str, str, str]]:
    return [(query["platform"].value, query["queue"].value, query["tier"].value, query["division"].value)]


###################
# Static Data API #
###################


# Champion

validate_champion_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_champion_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_champion(champion: Champion) -> List[Tuple]:
    keys = []
    try:
        keys.append((champion.platform.value, champion.version, champion.locale, _hash_included_data(champion.included_data), champion._data[ChampionData].id))
    except AttributeError:
        pass
    try:
        keys.append((champion.platform.value, champion.version, champion.locale, _hash_included_data(champion.included_data), champion._data[ChampionData].name))
    except AttributeError:
        pass
    return keys


def for_champion_query(query: Query) -> List[Tuple]:
    keys = []
    included_data_hash = _hash_included_data(query["includedData"])
    if "id" in query:
        keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, query["id"]))
    if "name" in query:
        keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, query["name"]))
    return keys


def for_many_champion_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    grouped_identifiers = []
    identifier_types = []
    if "ids" in query:
        grouped_identifiers.append(query["ids"])
        identifier_types.append(int)
    if "names" in query:
        grouped_identifiers.append(query["names"])
        identifier_types.append(str)
    for identifiers in zip(*grouped_identifiers):
        keys = []
        for identifier, identifier_type in zip(identifiers, identifier_types):
            try:
                identifier = identifier_type(identifier)
                keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, identifier))
            except ValueError as e:
                raise QueryValidationError from e
        yield keys


validate_champions_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_champions_query = Query. \
    has("platforms").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_champions(champions: Champions) -> List[Tuple[str, str, str, int]]:
    return [(champions.platform.value, champions.version, champions.locale, _hash_included_data(champions.included_data))]


def for_champions_query(query: Query) -> List[Tuple[str, str, str, int]]:
    included_data_hash = _hash_included_data(query["includedData"])
    return [(query["platform"].value, query["version"], query["locale"], included_data_hash)]


def for_many_champions_query(query: Query) -> Generator[List[Tuple[str, str, str, int, Union[int, str]]], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            yield [(platform.value, query["version"], query["locale"], included_data_hash)]
        except ValueError as e:
            raise QueryValidationError from e


# Item

validate_item_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_item_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_item(item: Item) -> List[Tuple]:
    keys = []
    try:
        keys.append((item.platform.value, item.version, item.locale, _hash_included_data(item.included_data), item._data[ItemData].id))
    except AttributeError:
        pass
    try:
        keys.append((item.platform.value, item.version, item.locale, _hash_included_data(item.included_data), item._data[ItemData].name))
    except AttributeError:
        pass
    return keys


def for_item_query(query: Query) -> List[Tuple]:
    keys = []
    included_data_hash = _hash_included_data(query["includedData"])
    if "id" in query:
        keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, query["id"]))
    if "name" in query:
        keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, query["name"]))
    return keys


def for_many_item_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    grouped_identifiers = []
    identifier_types = []
    if "ids" in query:
        grouped_identifiers.append(query["ids"])
        identifier_types.append(int)
    if "names" in query:
        grouped_identifiers.append(query["names"])
        identifier_types.append(str)
    for identifiers in zip(*grouped_identifiers):
        keys = []
        for identifier, identifier_type in zip(identifiers, identifier_types):
            try:
                identifier = identifier_type(identifier)
                keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, identifier))
            except ValueError as e:
                raise QueryValidationError from e
        yield keys


validate_items_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_items_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_items(items: Items) -> List[Tuple[str, str, str, int]]:
    return [(items.platform.value, items.version, items.locale, _hash_included_data(items.included_data))]


def for_items_query(query: Query) -> List[Tuple[str, str, str, int]]:
    included_data_hash = _hash_included_data(query["includedData"])
    return [(query["platform"].value, query["version"], query["locale"], included_data_hash)]


def for_many_items_query(query: Query) -> Generator[List[Tuple[str, str, str, int, Union[int, str]]], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            yield [(platform.value, query["version"], query["locale"], included_data_hash)]
        except ValueError as e:
            raise QueryValidationError from e


# Language

validate_languages_query = Query. \
    has("platform").as_(Platform)


validate_many_languages_query = Query. \
    has("platforms").as_(Iterable)


def for_languages(languages: Locales) -> List[str]:
    return [languages.platform.value]


def for_languages_query(query: Query) -> List[str]:
    return [query["platform"].value]


def for_many_languages_query(query: Query) -> Generator[List[str], None, None]:
    for platform in query["platforms"]:
        yield [platform.value]


validate_language_strings_query = Query. \
    has("platform").as_(Platform)


validate_many_language_strings_query = Query. \
    has("platforms").as_(Iterable)


def for_language_strings(languages: LanguageStrings) -> List[str]:
    return [languages.platform.value]


def for_language_strings_query(query: Query) -> List[str]:
    return [query["platform"].value]


def for_many_language_strings_query(query: Query) -> Generator[List[str], None, None]:
    for platform in query["platforms"]:
        yield [platform.value]


# Map

validate_map_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_map_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_map(map: Map) -> List[Tuple]:
    keys = []
    try:
        keys.append((map.platform.value, map.version, map.locale, map._data[MapData].id))
    except AttributeError:
        pass
    try:
        keys.append((map.platform.value, map.version, map.locale, map._data[MapData].name))
    except AttributeError:
        pass
    return keys


def for_map_query(query: Query) -> List[Tuple]:
    keys = []
    if "id" in query:
        keys.append((query["platform"].value, query["version"], query["locale"], query["id"]))
    if "name" in query:
        keys.append((query["platform"].value, query["version"], query["locale"], query["name"]))
    return keys


def for_many_map_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    grouped_identifiers = []
    identifier_types = []
    if "ids" in query:
        grouped_identifiers.append(query["ids"])
        identifier_types.append(int)
    if "names" in query:
        grouped_identifiers.append(query["names"])
        identifier_types.append(str)
    for identifiers in zip(*grouped_identifiers):
        keys = []
        for identifier, identifier_type in zip(identifiers, identifier_types):
            try:
                identifier = identifier_type(identifier)
                keys.append((query["platform"].value, query["version"], query["locale"], identifier))
            except ValueError as e:
                raise QueryValidationError from e
        yield keys


validate_maps_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


validate_many_maps_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


def for_maps(maps: Maps) -> List[Tuple[str, str, str]]:
    return [(maps.platform.value, maps.version, maps.locale)]


def for_maps_query(query: Query) -> List[Tuple[str, str, str]]:
    return [(query["platform"].value, query["version"], query["locale"])]


def for_many_maps_query(query: Query) -> Generator[List[Tuple[str, str, str]], None, None]:
    for platform in query["platforms"]:
        yield [(platform.value, query["version"], query["locale"])]


# Profile Icon

validate_profile_icons_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


validate_many_profile_icons_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str)


def for_profile_icons(profile_icon: ProfileIcons) -> List[Tuple[str, str, str]]:
    return [(Region(profile_icon.region).platform.value, profile_icon.version, profile_icon.locale)]


def for_profile_icons_query(query: Query) -> List[Tuple[str, str, str]]:
    return [(query["platform"].value, query["version"], query["locale"])]


def for_many_profile_icons_query(query: Query) -> Generator[List[Tuple[str, str, str]], None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield [(platform.value, query["version"], query["locale"])]
        except ValueError as e:
            raise QueryValidationError from e


validate_profile_icon_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("id").as_(int)


validate_many_profile_icon_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    has("ids").as_(Iterable)


def for_profile_icon(profile_icon: ProfileIcon) -> List[Tuple[str, str, str, int]]:
    return [(Region(profile_icon.region).platform.value, profile_icon.version, profile_icon.locale, profile_icon.id)]


def for_profile_icon_query(query: Query) -> List[Tuple[str, str, str, int]]:
    return [(query["platform"].value, query["version"], query["locale"], query["id"])]


def for_many_profile_icon_query(query: Query) -> Generator[List[Tuple[str, str, str, int]], None, None]:
    for id in query["ids"]:
        try:
            id = int(id)
            yield [(query["platform"].value, query["version"], query["locale"], id)]
        except ValueError as e:
            raise QueryValidationError from e


# Realm

validate_realms_query = Query. \
    has("platform").as_(Platform)


validate_many_realms_query = Query. \
    has("platforms").as_(Iterable)


def for_realms(realm: Realms) -> List[str]:
    return [(realm.platform.value)]


def for_realms_query(query: Query) -> List[str]:
    return [(query["platform"].value)]


def for_many_realms_query(query: Query) -> Generator[List[str], None, None]:
    for platform in query["platforms"]:
        yield [(platform.value)]


# Rune

validate_rune_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_rune_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_rune(rune: Rune) -> List[Tuple]:
    keys = []
    try:
        keys.append((rune.platform.value, rune.version, rune.locale, _hash_included_data(rune.included_data), rune._data[RuneData].id))
    except AttributeError:
        pass
    try:
        keys.append((rune.platform.value, rune.version, rune.locale, _hash_included_data(rune.included_data), rune._data[RuneData].name))
    except AttributeError:
        pass
    return keys


def for_rune_query(query: Query) -> List[Tuple]:
    keys = []
    included_data_hash = _hash_included_data(query["includedData"])
    if "id" in query:
        keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, query["id"]))
    if "name" in query:
        keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, query["name"]))
    return keys


def for_many_rune_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    grouped_identifiers = []
    identifier_types = []
    if "ids" in query:
        grouped_identifiers.append(query["ids"])
        identifier_types.append(int)
    if "names" in query:
        grouped_identifiers.append(query["names"])
        identifier_types.append(str)
    for identifiers in zip(*grouped_identifiers):
        keys = []
        for identifier, identifier_type in zip(identifiers, identifier_types):
            try:
                identifier = identifier_type(identifier)
                keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, identifier))
            except ValueError as e:
                raise QueryValidationError from e
        yield keys


validate_runes_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_runes_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_runes(runes: Runes) -> List[Tuple[str, str, str, int]]:
    return [(runes.platform.value, runes.version, runes.locale, _hash_included_data(runes.included_data))]


def for_runes_query(query: Query) -> List[Tuple[str, str, str, int]]:
    included_data_hash = _hash_included_data(query["includedData"])
    return [(query["platform"].value, query["version"], query["locale"], included_data_hash)]


def for_many_runes_query(query: Query) -> Generator[List[Tuple[str, str, str, int, Union[int, str]]], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            yield [(platform.value, query["version"], query["locale"], included_data_hash)]
        except ValueError as e:
            raise QueryValidationError from e


# Summoner Spell

validate_summoner_spell_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("id").as_(int).or_("name").as_(str)


validate_many_summoner_spell_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"}).also. \
    has("ids").as_(Iterable).or_("names").as_(Iterable)


def for_summoner_spell(summoner_spell: SummonerSpell) -> List[Tuple]:
    keys = []
    try:
        keys.append((summoner_spell.platform.value, summoner_spell.version, summoner_spell.locale, _hash_included_data(summoner_spell.included_data), summoner_spell._data[SummonerSpellData].id))
    except AttributeError:
        pass
    try:
        keys.append((summoner_spell.platform.value, summoner_spell.version, summoner_spell.locale, _hash_included_data(summoner_spell.included_data), summoner_spell._data[SummonerSpellData].name))
    except AttributeError:
        pass
    return keys


def for_summoner_spell_query(query: Query) -> List[Tuple]:
    keys = []
    included_data_hash = _hash_included_data(query["includedData"])
    if "id" in query:
        keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, query["id"]))
    if "name" in query:
        keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, query["name"]))
    return keys


def for_many_summoner_spell_query(query: Query) -> Generator[Tuple[str, str, str, int, Union[int, str]], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    grouped_identifiers = []
    identifier_types = []
    if "ids" in query:
        grouped_identifiers.append(query["ids"])
        identifier_types.append(int)
    if "names" in query:
        grouped_identifiers.append(query["names"])
        identifier_types.append(str)
    for identifiers in zip(*grouped_identifiers):
        keys = []
        for identifier, identifier_type in zip(identifiers, identifier_types):
            try:
                identifier = identifier_type(identifier)
                keys.append((query["platform"].value, query["version"], query["locale"], included_data_hash, identifier))
            except ValueError as e:
                raise QueryValidationError from e
        yield keys


validate_summoner_spells_query = Query. \
    has("platform").as_(Platform).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


validate_many_summoner_spells_query = Query. \
    has("platforms").as_(Iterable).also. \
    can_have("version").with_default(_get_default_version, supplies_type=str).also. \
    can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
    can_have("includedData").with_default({"all"})


def for_summoner_spells(summoner_spells: SummonerSpells) -> List[Tuple[str, str, str, int]]:
    return [(summoner_spells.platform.value, summoner_spells.version, summoner_spells.locale, _hash_included_data(summoner_spells.included_data))]


def for_summoner_spells_query(query: Query) -> List[Tuple[str, str, str, int]]:
    included_data_hash = _hash_included_data(query["includedData"])
    return [(query["platform"].value, query["version"], query["locale"], included_data_hash)]


def for_many_summoner_spells_query(query: Query) -> Generator[List[Tuple[str, str, str, int, Union[int, str]]], None, None]:
    included_data_hash = _hash_included_data(query["includedData"])
    for platform in query["platforms"]:
        try:
            yield [(platform.value, query["version"], query["locale"], included_data_hash)]
        except ValueError as e:
            raise QueryValidationError from e


# Version

validate_versions_query = Query. \
    has("platform").as_(Platform)


validate_many_versions_query = Query. \
    has("platforms").as_(Iterable)


def for_versions(versions: Versions) -> List[str]:
    return [versions.platform.value]


def for_versions_query(query: Query) -> List[str]:
    return [query["platform"].value]


def for_many_versions_query(query: Query) -> Generator[List[str], None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield [platform.value]
        except ValueError as e:
            raise QueryValidationError from e


##############
# Status API #
##############

validate_shard_status_query = Query. \
    has("platform").as_(Platform)


validate_many_shard_status_query = Query. \
    has("platforms").as_(Iterable)


def for_shard_status(shard_status: ShardStatus) -> List[str]:
    return [shard_status.platform.value]


def for_shard_status_query(query: Query) -> List[str]:
    return [query["platform"].value]


def for_many_shard_status_query(query: Query) -> Generator[List[str], None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield [platform.value]
        except ValueError as e:
            raise QueryValidationError from e


#############
# Match API #
#############


validate_match_query = Query. \
    has("platform").as_(Platform).also. \
    has("id").as_(int)


validate_many_match_query = Query. \
    has("platform").as_(Platform).also. \
    has("ids").as_(Iterable)


def for_match(match: Match) -> List[Tuple[str, int]]:
    return [(match.platform.value, match.id)]


def for_match_query(query: Query) -> List[Tuple[str, int]]:
    return [(query["platform"].value, query["id"])]


def for_many_match_query(query: Query) -> Generator[List[Tuple[str, int]], None, None]:
    for id in query["ids"]:
        try:
            id = int(id)
            yield [(query["platform"].value, id)]
        except ValueError as e:
            raise QueryValidationError from e

validate_match_timeline_query = Query. \
    has("platform").as_(Platform).also. \
    has("id").as_(int)


validate_many_match_timeline_query = Query. \
    has("platform").as_(Platform).also. \
    has("ids").as_(Iterable)


def for_match_timeline(timeline: Timeline) -> List[Tuple[str, int]]:
    return [(timeline.platform.value, timeline.id)]


def for_match_timeline_query(query: Query) -> List[Tuple[str, int]]:
    return [(query["platform"].value, query["id"])]


def for_many_match_timeline_query(query: Query) -> Generator[List[Tuple[str, int]], None, None]:
    for id in query["ids"]:
        try:
            id = int(id)
            yield [(query["platform"].value, id)]
        except ValueError as e:
            raise QueryValidationError from e


#################
# Spectator API #
#################

validate_current_match_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.id").as_(str)


validate_many_current_match_query = Query. \
    has("platform").as_(Platform).also. \
    has("summoner.ids").as_(Iterable)


def for_current_match(current_match_info: CurrentMatch) -> List[Tuple[str, str]]:
    # Reach into the data for the summoner ids so we don't create the Summoner objects
    # This stores the current match for every summoner in the match, so if a different summoner is
    #  requested, the match isn't pulled a second time.
    return [(current_match_info.platform.value, participant._data[CurrentGameParticipantData].summonerId) for participant in current_match_info.participants] + [(current_match_info.platform.value, current_match_info.id)]


def for_current_match_query(query: Query) -> List[Tuple[str, str]]:
    return [(query["platform"].value, query["summoner.id"])]


def for_many_current_match_query(query: Query) -> Generator[List[Tuple[str, str]], None, None]:
    for summoner_id in query["summoner.ids"]:
        try:
            summoner_id = int(summoner_id)
            yield [(query["platform"].value, summoner_id)]
        except ValueError as e:
            raise QueryValidationError from e


validate_featured_matches_query = Query. \
    has("platform").as_(Platform)


validate_many_featured_matches_query = Query. \
    has("platforms").as_(Iterable)


def for_featured_matches(featured_matches: FeaturedMatches) -> List[str]:
    return [featured_matches.platform]


def for_featured_matches_query(query: Query) -> List[str]:
    return [query["platform"].value]


def for_many_featured_matches_query(query: Query) -> Generator[List[str], None, None]:
    for platform in query["platforms"]:
        try:
            platform = Platform(platform)
            yield [platform.value]
        except ValueError as e:
            raise QueryValidationError from e


################
# Summoner API #
################


validate_summoner_query = Query. \
    has("platform").as_(Platform).also. \
    has("id").as_(str).or_("accountId").as_(str).or_("name").as_(str).or_("puuid").as_(str)


validate_many_summoner_query = Query. \
    has("platform").as_(Platform).also. \
    has("ids").as_(Iterable).or_("accountIds").as_(Iterable).or_("names").as_(Iterable).or_("puuids").as_(Iterable)


def for_summoner(summoner: Summoner) -> List[Tuple]:
    keys = []
    try:
        keys.append((summoner.platform.value, "id", summoner._data[SummonerData].id))
    except AttributeError:
        pass
    try:
        keys.append((summoner.platform.value, "name", summoner._data[SummonerData].name))
    except AttributeError:
        pass
    try:
        keys.append((summoner.platform.value, "accountId", summoner._data[SummonerData].accountId))
    except AttributeError:
        pass
    try:
        keys.append((summoner.platform.value, "puuid", summoner._data[SummonerData].puuid))
    except AttributeError:
        pass
    return keys


def for_summoner_query(query: Query) -> List[Tuple]:
    keys = []
    if "id" in query:
        keys.append((query["platform"].value, "id", query["id"]))
    if "name" in query:
        keys.append((query["platform"].value, "name", query["name"]))
    if "accountId" in query:
        keys.append((query["platform"].value, "accountId", query["accountId"]))
    if "puuid" in query:
        keys.append((query["platform"].value, "puuid", query["puuid"]))
    return keys


def for_many_summoner_query(query: Query) -> Generator[List[Tuple], None, None]:
    grouped_identifiers = []
    identifier_types = []
    if "ids" in query:
        grouped_identifiers.append(query["ids"])
        identifier_types.append(str)
    elif "accountIds" in query:
        grouped_identifiers.append(query["accountIds"])
        identifier_types.append(str)
    elif "puuids" in query:
        grouped_identifiers.append(query["puuids"])
        identifier_types.append(str)
    elif "names" in query:
        grouped_identifiers.append(query["names"])
        identifier_types.append(str)
    for identifiers in zip(*grouped_identifiers):
        keys = []
        for identifier, identifier_type in zip(identifiers, identifier_types):
            try:
                identifier = identifier_type(identifier)
                keys.append((query["platform"].value, identifier))
            except ValueError as e:
                raise QueryValidationError from e
        yield keys



from typing import Type, TypeVar

from datapipelines import DataTransformer, PipelineContext

from ..core.thirdpartycode import VerificationString, VerificationStringData
from ..dto.thirdpartycode import VerificationStringDto

T = TypeVar("T")
F = TypeVar("F")


class ThirdPartyCodeTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # Dto to Data

    @transform.register(VerificationStringDto, VerificationStringData)
    def verification_string_dto_to_data(self, value: VerificationStringDto, context: PipelineContext = None) -> VerificationStringData:
        return VerificationStringData(**value)

    # Data to Core

    #@transform.register(VerificationStringData, VerificationString)
    def verification_string_data_to_core(self, value: VerificationStringData, context: PipelineContext = None) -> VerificationString:
        return VerificationString.from_data(value)



from typing import TypeVar, Type, Dict, Union, List, Mapping
from django.conf import settings
import copy
from logging import getLogger
LOGGER = getLogger(__name__)

expire_index = {
    "cr":"ChampionRotationData",
    "rl":"Realms",
    "v":"Versions",
    "c":"Champion",
    "c+":"Champions",
    "r":"Rune",
    "r+":"Runes",
    "i":"Item",
    "i+":"Items",
    "ss":"SummonerSpell",
    "ss+":"SummonerSpells",
    "mp":"Map",
    "mp+":"Maps",
    "pi":"ProfileIcon",
    "pi+":"ProfileIcons",
    "ls":"Locales",
    "ls+":"LanguageStrings",
    "cm":"ChampionMastery",
    "cm+":"ChampionMasteries",
    "lse":"LeagueSummonerEntries",
    "l":"League",
    "cl":"ChallengerLeague",
    "gl":"GrandmasterLeague",
    "ml":"MasterLeague",
    "m":"Match",
    "t":"Timeline",
    "s":"Summoner",
    "shs":"ShardStatus",
    "cg":"CurrentMatch",
    "fg":"FeaturedMatches",
    "rl-":"RealmDto",
    "v-":"VersionListDto",
    "c-":"ChampionDto",
    "c+-":"ChampionListDto",
    "r-":"RuneDto",
    "r+-":"RuneListDto",
    "i-":"ItemDto",
    "i+-":"ItemListDto",
    "ss-":"SummonerSpellDto",
    "ss+-":"SummonerSpellListDto",
    "mp-":"MapDto",
    "mp+-":"MapListDto",
    "pi-":"ProfileIconDetailsDto",
    "pi+-":"ProfileIconDataDto",
    "ls-":"LanguagesDto",
    "ls+-":"LanguageStringsDto",
    "cr-":"ChampionRotationDto",
    "cm-":"ChampionMasteryDto",
    "cm+-":"ChampionMasteryListDto",
    "cl-":"ChallengerLeagueListDto",
    "gl-":"GrandmasterLeagueListDto",
    "ml-":"MasterLeagueListDto",
    "m-":"MatchDto",
    "t-":"TimelineDto",
    "s-":"SummonerDto",
    "shs-":"ShardStatusDto",
    "cg-":"CurrentGameInfoDto",
    "fg-":"FeaturedGamesDto",
    "p-":"PatchListDto",
    "*-": "*-",
    "*+": "*+",
    "**": "**"
}

error_handler_map = {
    "t": {
        "strategy": "throw"
    },
    "^e": {
        "strategy": "exponential_backoff",
        1 : "initial_backoff",
        2 : "backoff_factor",
        3 : "max_attempts",
    },
    "r": {
        "strategy": "retry_from_headers",
        1 : "max_attempts",
    }
}


def create_expiration(mape) -> Mapping[type, dict]:
    expirations = {}
    last = False
    for time, types in mape.items():
        if last:
            raise AttributeError("key of type '*' is assigned but not last of object 'expirations_map'")
        for typ in types:
            if typ[0] == "*" and typ[-1] in ["*","-","+"] and len(typ) == 2:
                for mapkey, mapobj in expire_index.items():
                    key_to_add = None
                    if typ == "*-" and mapobj not in expirations.keys() and mapobj.endswith('Dto'):
                        key_to_add = mapobj
                    elif typ == "*+" and mapobj not in expirations.keys() and not mapobj.endswith('Dto'):
                        key_to_add = mapobj
                    elif typ == "**":
                        key_to_add = mapobj
                    if key_to_add is not None and key_to_add not in ["**","*-","*+"]:
                        expirations[mapobj] = time
                last = True
            elif expire_index[typ] not in expirations.keys():
                    expirations[expire_index[typ]] = time
            else:
                if typ[0] == "*":
                    raise KeyError("'"+typ+"' in object 'expirations_map'")
                raise AttributeError("Attempted duplicate key '"+typ+"' in object 'expirations_map'")
    return expirations


def create_handler(mape) -> Mapping[type, type]:
    handler = {}
    for code, values in mape.items():
        if code not in handler.keys():
            handler[code]= {}
        else:
            raise KeyError("Attempted duplicate key '"+code+"' in object 'riotapi_request_error_handling'")
        if code == "429":
            for limit_type, inner_values in values.items():
                strategy = error_handler_map[inner_values[0]]
                handler[code][limit_type] = {}
                for ind in range(len(inner_values)):
                    if ind == 0:
                        handler[code][limit_type]["strategy"] = strategy["strategy"]
                    else:
                        handler[code][limit_type][strategy[ind]] = inner_values[ind]
        else:
            for ind in range(len(values)):
                strategy = error_handler_map[values[0]]
                if ind == 0:
                    handler[code]["strategy"] = strategy["strategy"]
                else:
                    handler[code][strategy[ind]] = values[ind]
    return handler


def get_cass_settings():
    cass = {
        "global" : getattr(settings, "CASSIOPEIA_GLOBAL", None),
        "pipeline" : getattr(settings, "CASSIOPEIA_PIPELINE", None),
        "logging" : getattr(settings, "CASSIOPEIA_LOGGING", None),
        "plugins": getattr(settings, "CASSIOPEIA_PLUGINS", None)
    }
    for key, config in copy.deepcopy(cass).items():
        if key == "pipeline" and config is None:
            raise AttributeError("'pipeline' settings in django-cassiopeia is obligatory")
        elif config is None:
            LOGGER.warning(f"[Traceback: django-cassiopeia > settings] WARNING: '{key}' is not set, using default settings")
            del cass[key]
    for store_name, config in copy.deepcopy(cass["pipeline"]).items():
        if store_name == "DjangoCache":
            cass["pipeline"][store_name] = settings.CASSIOPEIA_DJANGO_CACHES
        elif store_name == "RiotAPI":
            cass["pipeline"][store_name]["api_key"] = settings.RIOT_API_KEY
            cass["pipeline"][store_name]["limiting_share"] = settings.CASSIOPEIA_LIMITING_SHARE
            try:
                handler = settings.CASSIOPEIA_API_ERROR_HANDLING
                cass["pipeline"][store_name]["request_error_handling"] = create_handler(handler)
            except AttributeError:
                LOGGER.warning("[Traceback: django-cassiopeia > settings] Warning: 'riotapi_request_error_handling' is not set, using default settings")
                LOGGER.warning("[Traceback: django-cassiopeia > settings] Warning: 'riotapi_request_error_handling' has changed its syntax since version 1.1.0."+
                    "No longer accepts 'CASSIOPEIA_RIOT_API_ERROR_HANDLING', a shorter version is replaced on 'CASSIOPEIA_API_ERROR_HANDLING', see documentation.")
    for store_name, config in copy.deepcopy(cass["pipeline"]).items():
        if store_name.lower() == "djangocache":
            for ind in range(len(config)):
                if "expirations_map" in config[ind].keys():
                    cass["pipeline"][store_name][ind]["expirations"] = create_expiration(cass["pipeline"][store_name][ind].pop("expirations_map"))
        else:
            if "expirations_map" in config.keys():
                cass["pipeline"][store_name]["expirations"] = create_expiration(cass["pipeline"][store_name].pop("expirations_map"))
    return cass


from typing import Type, TypeVar
from copy import deepcopy

from datapipelines import DataTransformer, PipelineContext

from ..core.champion import ChampionRotationData, ChampionRotation
from ..dto.champion import ChampionRotationDto

T = TypeVar("T")
F = TypeVar("F")


class ChampionTransformer(DataTransformer):
    @DataTransformer.dispatch
    def transform(self, target_type: Type[T], value: F, context: PipelineContext = None) -> T:
        pass

    # Dto to Data

    @transform.register(ChampionRotationDto, ChampionRotationData)
    def champion_rotation_dto_to_data(self, value: ChampionRotationDto, context: PipelineContext = None) -> ChampionRotationData:
        return ChampionRotationData(**value)

    # Data to Core

    @transform.register(ChampionRotationData, ChampionRotation)
    def champion_rotation_data_to_core(self, value: ChampionRotationData, context: PipelineContext = None) -> ChampionRotation:
        return ChampionRotation.from_data(value, loaded_groups={ChampionRotationData})



from typing import Union

from merakicommons.cache import lazy_property

from ..data import Region, Platform
from .common import CoreData, CassiopeiaGhost, provide_default_region, ghost_load_on
from .summoner import Summoner


##############
# Data Types #
##############


class VerificationStringData(CoreData):
    _renamed = {}


##############
# Core Types #
##############


class VerificationString(CassiopeiaGhost):
    _data_types = {VerificationStringData}

    @provide_default_region
    def __init__(self, summoner: Summoner, region: Union[Region, str]):
        self.__summoner = summoner
        kwargs = {"region": region}
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "summoner.id": self.summoner.id}

    @classmethod
    @provide_default_region
    def __get_query_from_kwargs__(cls, *, summoner: Summoner, region: Union[Region, str]) -> dict:
        return {"region": region, "summoner.id": summoner.id}

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[VerificationStringData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @CassiopeiaGhost.property(VerificationStringData)
    @ghost_load_on
    def string(self) -> str:
        return self._data[VerificationStringData].string

    @property
    def summoner(self):
        return self.__summoner


from typing import Type, TypeVar, MutableMapping, Any, Iterable

from datapipelines import DataSource, PipelineContext, Query, NotFoundError, validate_query
from .common import RiotAPIService, APINotFoundError
from ...data import Platform
from ...dto.summoner import SummonerDto
from ..uniquekeys import convert_region_to_platform

T = TypeVar("T")


class SummonerAPI(RiotAPIService):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    _validate_get_summoner_query = Query. \
        has("id").as_(str). \
        or_("accountId").as_(str). \
        or_("puuid").as_(str). \
        or_("name").as_(str).also. \
        has("platform").as_(Platform)

    @get.register(SummonerDto)
    @validate_query(_validate_get_summoner_query, convert_region_to_platform)
    def get_summoner(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> SummonerDto:
        if "id" in query:
            url = "https://{platform}.api.riotgames.com/lol/summoner/v4/summoners/{summonerId}".format(platform=query["platform"].value.lower(), summonerId=query["id"])
            endpoint = "summoners/summonerId"
        elif "accountId" in query:
            url = "https://{platform}.api.riotgames.com/lol/summoner/v4/summoners/by-account/{accountId}".format(platform=query["platform"].value.lower(), accountId=query["accountId"])
            endpoint = "summoners/by-account/accountId"
        elif "name" in query:
            url = "https://{platform}.api.riotgames.com/lol/summoner/v4/summoners/by-name/{name}".format(platform=query["platform"].value.lower(), name=query["name"].replace(" ", "")).encode("utf-8")
            endpoint = "summoners/by-name/name"
        elif "puuid" in query:
            url = url = "https://{platform}.api.riotgames.com/lol/summoner/v4/summoners/by-puuid/{puuid}".format(platform=query["platform"].value.lower(), puuid=query["puuid"])
            endpoint = "summoners/by-puuid/puuid"
        else:
            endpoint = ""

        try:
            app_limiter, method_limiter = self._get_rate_limiter(query["platform"], endpoint)
            data = self._get(url, {}, app_limiter=app_limiter, method_limiter=method_limiter)
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        return SummonerDto(**data)



from typing import List, Union, Set

from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable, SearchableList

from ...data import Resource, Region, Platform, GameMode
from ..common import CoreData, CoreDataList, CassiopeiaObject, CassiopeiaGhost, CassiopeiaLazyList, get_latest_version, provide_default_region, ghost_load_on
from .common import ImageData, Image, Sprite
from ...dto.staticdata import summonerspell as dto


##############
# Data Types #
##############


class SummonerSpellListData(CoreDataList):
    _dto_type = dto.SummonerSpellListDto
    _renamed = {"included_data": "includedData"}


class SpellVarsData(CoreData):
    _renamed = {"dyn": "dynamic", "coeff": "coefficients"}


class LevelTipData(CoreData):
    _renamed = {"effect": "effects", "label": "keywords"}


class SummonerSpellData(CoreData):
    _dto_type = dto.SummonerSpellDto
    _renamed = {"maxrank": "maxRank", "cooldown": "cooldowns", "cost": "costs", "effect": "effects", "costType": "resource", "included_data": "includedData"}

    def __call__(self, **kwargs):
        if "vars" in kwargs:
            self.variables = [SpellVarsData(**v) for v in kwargs.pop("vars")]
        if "altimages" in kwargs:
            self.alternativeImages = [ImageData(**alt) for alt in kwargs.pop("altimages")]
        if "image" in kwargs:
            self.image = ImageData(**kwargs.pop("image"))
        super().__call__(**kwargs)
        return self


##############
# Core Types #
##############


class SummonerSpells(CassiopeiaLazyList):
    _data_types = {SummonerSpellListData}

    @provide_default_region
    def __init__(self, *, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if included_data is None:
            included_data = {"all"}
        if locale is None and region is not None:
            locale = Region(region).default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if version:
            kwargs["version"] = version
        CassiopeiaObject.__init__(self, **kwargs)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[SummonerSpellListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def version(self) -> str:
        try:
            return self._data[SummonerSpellListData].version
        except AttributeError:
            version = get_latest_version(region=self.region, endpoint="summoner")
            self(version=version)
            return self._data[SummonerSpellListData].version

    @property
    def locale(self) -> str:
        """The locale for this champion."""
        return self._data[SummonerSpellListData].locale

    @property
    def included_data(self) -> Set[str]:
        """A set of tags to return additonal information for this champion when it's loaded."""
        return self._data[SummonerSpellListData].includedData


@searchable({str: ["key"]})
class SpellVars(CassiopeiaObject):
    _data_types = {SpellVarsData}

    @property
    def ranks_with(self) -> str:
        """Well, we don't know what this one is. let us know if you figure it out."""
        return self._data[SpellVarsData].ranksWith

    @property
    def dynamic(self) -> str:
        """Well, we don't know what this one is. let us know if you figure it out."""
        return self._data[SpellVarsData].dynamic

    @property
    def link(self) -> str:
        """Stat this spell scales from."""
        return self._data[SpellVarsData].link

    @lazy_property
    def coefficients(self) -> List[float]:
        """The scaling coefficients for this spell."""
        return SearchableList(self._data[SpellVarsData].coefficients)

    @property
    def key(self) -> str:
        """Well, we don't know what this one is. let us know if you figure it out."""
        return self._data[SpellVarsData].key


@searchable({str: ["name", "key", "keywords"]})
class SummonerSpell(CassiopeiaGhost):
    _data_types = {SummonerSpellData}

    @provide_default_region
    def __init__(self, *, id: int = None, name: str = None, region: Union[Region, str] = None, version: str = None, locale: str = None, included_data: Set[str] = None):
        if included_data is None:
            included_data = {"all"}
        if locale is None and region is not None:
            locale = Region(region).default_locale
        kwargs = {"region": region, "included_data": included_data, "locale": locale}
        if id is not None:
            kwargs["id"] = id
        if name is not None:
            kwargs["name"] = name
        if version is not None:
            kwargs["version"] = version
        super().__init__(**kwargs)

    def __get_query__(self):
        query = {"region": self.region, "platform": self.platform, "version": self.version, "locale": self.locale, "includedData": self.included_data}
        if hasattr(self._data[SummonerSpellData], "id"):
            query["id"] = self._data[SummonerSpellData].id
        if hasattr(self._data[SummonerSpellData], "name"):
            query["name"] = self._data[SummonerSpellData].name
        return query

    def __eq__(self, other: "SummonerSpell"):
        if not isinstance(other, SummonerSpell) or self.region != other.region:
            return False
        s = {}
        o = {}
        if hasattr(self._data[SummonerSpellData], "id"): s["id"] = self.id
        if hasattr(other._data[SummonerSpellData], "id"): o["id"] = other.id
        if hasattr(self._data[SummonerSpellData], "name"): s["name"] = self.name
        if hasattr(other._data[SummonerSpellData], "name"): o["name"] = other.name
        if any(s.get(key, "s") == o.get(key, "o") for key in s):
            return True
        else:
            return self.id == other.id

    def __str__(self):
        region = self.region
        id_ = "?"
        name = "?"
        if hasattr(self._data[SummonerSpellData], "id"):
            id_ = self.id
        if hasattr(self._data[SummonerSpellData], "name"):
            name = self.name
        return "SummonerSpell(name='{name}', id={id_}, region='{region}')".format(name=name, id_=id_, region=region.value)

    __hash__ = CassiopeiaGhost.__hash__

    @lazy_property
    def region(self) -> Region:
        """The region for this summoner spell."""
        return Region(self._data[SummonerSpellData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this summoner spell."""
        return self.region.platform

    @property
    def version(self) -> str:
        """The version for this summoner spell."""
        try:
            return self._data[SummonerSpellData].version
        except AttributeError:
            version = get_latest_version(region=self.region, endpoint="summoner")
            self(version=version)
            return self._data[SummonerSpellData].version

    @property
    def locale(self) -> str:
        """The locale for this summoner spell."""
        return self._data[SummonerSpellData].locale or self.region.default_locale

    @property
    def included_data(self) -> Set[str]:
        """The data to included in the query for this summoner spell."""
        return self._data[SummonerSpellData].includedData

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    @lazy
    def modes(self) -> List[GameMode]:
        return SearchableList([GameMode(mode) for mode in self._data[SummonerSpellData].modes])

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    @lazy
    def variables(self) -> List[SpellVars]:
        """Contains spell data."""
        return SearchableList(SpellVars(v) for v in self._data[SummonerSpellData].variables)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    @lazy
    def resource(self) -> Resource:
        """The resource consumed when using this spell."""
        return Resource(self._data[SummonerSpellData].resource)

    @lazy_property
    def sprite(self) -> Sprite:
        return self.image.sprite_info

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    def sanitized_description(self) -> str:
        """The spell's sanitized description."""
        return self._data[SummonerSpellData].sanitized_description

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    def sanitized_tooltip(self) -> str:
        """The spell's sanitized tooltip."""
        return self._data[SummonerSpellData].sanitizedTooltip

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    @lazy
    def effects(self) -> List[List[float]]:
        """The level-by-level replacements for {{ e# }} tags in other values."""
        return SearchableList(self._data[SummonerSpellData].effects)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    def tooltip(self) -> str:
        """The spell's tooltip."""
        return self._data[SummonerSpellData].tooltip

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    def max_rank(self) -> int:
        """The maximum rank this spell can attain."""
        return self._data[SummonerSpellData].maxRank

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    @lazy
    def range(self) -> List[Union[int, str]]:
        """The maximum range of this spell. `self` if it has no range."""
        return SearchableList(self._data[SummonerSpellData].range)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    @lazy
    def cooldowns(self) -> List[float]:
        """The cooldowns of this spell (per level)."""
        return SearchableList(self._data[SummonerSpellData].cooldowns)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    @lazy
    def costs(self) -> List[int]:
        """The resource costs of this spell (per level)."""
        return SearchableList(self._data[SummonerSpellData].costs)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    def key(self) -> str:
        """The spell's key."""
        return self._data[SummonerSpellData].key

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    def description(self) -> str:
        """The spell's description."""
        return self._data[SummonerSpellData].description

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    @lazy
    def alternative_images(self) -> List[Image]:
        """The alternative images for this spell. These won't exist after patch NN, when Riot standardized all images."""
        return SearchableList(Image.from_data(alt) for alt in self._data[SummonerSpellData].alternativeImages)

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    def id(self) -> int:
        """The spell's id."""
        return self._data[SummonerSpellData].id

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    def name(self) -> str:
        """The spell's name."""
        return self._data[SummonerSpellData].name

    @CassiopeiaGhost.property(SummonerSpellData)
    @ghost_load_on
    @lazy
    def image(self) -> Image:
        image = Image.from_data(self._data[SummonerSpellData].image)
        image(version=self.version)
        return image

    @lazy_property
    def sprite(self) -> Sprite:
        return self.image.spriteInfo



from typing import Dict, Union

from merakicommons.cache import lazy_property
from merakicommons.container import searchable

from ...data import Region, Platform
from ..common import CoreData, CassiopeiaGhost, provide_default_region, ghost_load_on
from ...dto.staticdata import realm as dto


##############
# Data Types #
##############


class RealmData(CoreData):
    _dto_type = dto.RealmDto
    _renamed = {"lg": "legacyMode", "dd": "latestDataDragon", "l": "language", "n": "latestVersions",
                "profileiconmax": "maxProfileIconId", "v": "version", "css": "cssVersion"}


##############
# Core Types #
##############


@searchable({})
class Realms(CassiopeiaGhost):
    _data_types = {RealmData}

    @provide_default_region
    def __init__(self, region: Union[Region, str] = None):
        kwargs = {"region": region}
        super().__init__(**kwargs)

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform}

    @lazy_property
    def region(self) -> Region:
        """The region for this realm."""
        return Region(self._data[RealmData].region)

    @lazy_property
    def platform(self) -> Platform:
        """The platform for this realm."""
        return self.region.platform

    @lazy_property
    def locale(self) -> Platform:
        """The locale for this realm."""
        return self._data[RealmData].locale

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on
    def version(self) -> str:
        return self._data[RealmData].version

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on
    def language(self) -> str:
        """Default language for this realm."""
        return self._data[RealmData].language

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on
    def latest_versions(self) -> Dict[str, str]:
        """Latest changed version for each data type listed."""
        return self._data[RealmData].latestVersions

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on
    def legacy_mode(self) -> str:
        return self._data[RealmData].legacyMode

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on
    def latest_data_dragon(self) -> str:
        return self._data[RealmData].latestDataDragon

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on
    def language(self) -> str:
        return self._data[RealmData].language

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on
    def max_profile_icon_id(self) -> int:
        return self._data[RealmData].maxProfileIconId

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on
    def store(self) -> str:
        return self._data[RealmData].store

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on
    def cdn(self) -> str:
        return self._data[RealmData].cdn

    @CassiopeiaGhost.property(RealmData)
    @ghost_load_on
    def css_version(self) -> str:
        return self._data[RealmData].css_version



from typing import Union

from merakicommons.cache import lazy_property

from ...data import Region, Platform
from ...dto.staticdata.version import VersionListDto
from ..common import CoreDataList, CassiopeiaObject, CassiopeiaLazyList, provide_default_region


class VersionListData(CoreDataList):
    _dto_type = VersionListDto
    _renamed = {}


class Versions(CassiopeiaLazyList):
    _data_types = {VersionListData}

    @provide_default_region
    def __init__(self, *, region: Union[Region, str] = None):
        kwargs = {"region": region}
        CassiopeiaObject.__init__(self, **kwargs)

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[VersionListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform



from abc import abstractmethod
from typing import MutableMapping, Any, Union, TypeVar, Iterable, Type
try:
    import ujson as json
except ImportError:
    import json

from datapipelines import DataSource, PipelineContext

from ..common import HTTPClient, HTTPError, Curl
from ..riotapi.common import APIForbiddenError, APINotFoundError, APIRequestError, APIError, _ERROR_CODES
from ...dto.staticdata.realm import RealmDto


def _get_latest_version(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    pipeline = context[PipelineContext.Keys.PIPELINE]
    realms = pipeline.get(RealmDto, {"platform": query["platform"]})
    return realms["v"]


def _get_default_locale(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    return query["platform"].default_locale


T = TypeVar("T")


class KernelSource(DataSource):
    def __init__(self, server_url: str, port: int, http_client: HTTPClient = None):
        self._server_url = server_url
        self._port = port
        if http_client is None:
            self._client = HTTPClient()
        else:
            self._client = http_client

    def _get(self, endpoint: str, parameters: MutableMapping[str, Any] = None, connection: Curl = None) -> Union[dict, list, Any]:
        url = f"{self._server_url}:{self._port}/{endpoint}"
        try:
            result, headers = self._client.get(url=url, parameters=parameters, connection=connection)

            # Ensure compatibility with both Curl and Requests based clients
            if isinstance(result, bytes):
                result = json.loads(result.decode())
            elif isinstance(result, str):
                result = json.loads(result)
            
            if not isinstance(result, dict):
                raise ValueError("Unexpected type returned from HTTPClient: {}".format(type(result)))
        except HTTPError as error:
            # The error handlers didn't work, so raise an appropriate error.
            new_error_type = _ERROR_CODES[error.code]
            if new_error_type is RuntimeError:
                new_error = RuntimeError("Encountered an HTTP error code {code} with message \"{message}\" which should have already been handled. Report this to the Cassiopeia team.".format(code=error.code, message=str(error)))
            elif new_error_type is APIError:
                new_error = APIError("Kernel experienced an internal error on the request. You may want to retry the request after a short wait or continue without the result. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            elif new_error_type is APINotFoundError:
                new_error = APINotFoundError("Kernel returned a NOT FOUND error for the request. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            elif new_error_type is APIRequestError:
                new_error = APIRequestError("Kernel returned an error on the request. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            elif new_error_type is APIForbiddenError:
                new_error = APIForbiddenError("Kernel returned a FORBIDDEN error for the request. The received error was {code}: \"{message}\"".format(code=error.code, message=str(error)), error.code)
            else:
                new_error = new_error_type(str(error))

            raise new_error from error
        return result

    @abstractmethod
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @abstractmethod
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass



from functools import wraps
from typing import Callable, Any, TypeVar
from collections import defaultdict
from django.core.cache import caches
import datetime

from logging import getLogger

LOGGER = getLogger(__name__)
T = TypeVar("T")

class DjangoCacheBackend(object):
    def __init__(self, alias: str = None, logs_enabled: bool = True) -> None:
        self._alias = alias
        self._cache = caches[alias]
        self._logs_enabled = logs_enabled

    def put(self, key: Any, value: Any, timeout: int = -1) -> None:
        if timeout != 0:
            if timeout == -1:
                timeout = None
            self._cache.set(key, value, timeout)
            if self._logs_enabled:
                LOGGER.warn(f"[Traceback: django_cassiopeia > {self._alias}] PUT: {key}")

    def put_many(self, pairs: Any, timeout: int = -1) -> None:
        if timeout != 0:
            if timeout == -1:
                timeout = None
            self._cache.set_many(pairs, timeout)
            if self._logs_enabled:
                LOGGER.warn(f"[Traceback: django_cassiopeia > {self._alias}] PUT-MANY: {[key for key in pairs.keys()]}")

    def get(self, key: Any) -> Any:
        item = self._cache.get(key)
        if item is None:
            raise KeyError
        if self._logs_enabled:
            LOGGER.warn(f"[Traceback: django_cassiopeia > {self._alias}] GET: {key}")
        return item

    # This varies from the cache backend you use, might implement
    def get_all(self, type: Any):
        raise NotImplementedError

    # This should be rarely used, no delete() implemented in cassiopeia
    # You need to manually enter the key
    # For django you can just simply delete it by entering your cache backend server/db
    def delete(self, key: Any) -> None:
        self._cache.delete(key)

    # Again, cass not implemented, you can do it on your django backend server/db
    # but option is here if you want to make life harder
    def contains(self, key: Any) -> bool:
        item = self._cache.get(key)
        if item is None:
            return False
        return True
    
    # This function is really not needed.
    # It is more than a helper func if you're not familiar with django cache.
    # Django way of clearing:
    # terminal/cmd: `python manage.py shell`
    # >>> from django.core.cache import caches
    # >>> caches[name_of_your_cache].clear()
    def clear(self, pref: str = None):
        if pref is None:
            self._cache.clear()
        else:
            self._cache.clear(prefix=pref)

    # No need to expire, Django handles it.
    # For filebased/database backends, it should have its own culling system.
    def expire(self, type: Any = None):
        pass




from typing import Iterable, Set, Dict
import os

from datapipelines import CompositeDataSource
from .common import RiotAPIService, RiotAPIRateLimiter


def _default_services(api_key: str, limiting_share: float = 1.0, request_error_handling: Dict = None) -> Set[RiotAPIService]:
    from ..common import HTTPClient
    from ..image import ImageDataSource
    from .champion import ChampionAPI
    from .summoner import SummonerAPI
    from .championmastery import ChampionMasteryAPI
    from .match import MatchAPI
    from .spectator import SpectatorAPI
    from .status import StatusAPI
    from .leagues import LeaguesAPI
    from .thirdpartycode import ThirdPartyCodeAPI
    from ...data import Platform

    app_rate_limiter = {platform: RiotAPIRateLimiter(limiting_share=limiting_share) for platform in Platform}

    client = HTTPClient()
    services = {
        ImageDataSource(client),
        ChampionAPI(api_key, app_rate_limiter=app_rate_limiter, request_error_handling=request_error_handling, http_client=client),
        SummonerAPI(api_key, app_rate_limiter=app_rate_limiter, request_error_handling=request_error_handling, http_client=client),
        ChampionMasteryAPI(api_key, app_rate_limiter=app_rate_limiter, request_error_handling=request_error_handling, http_client=client),
        MatchAPI(api_key, app_rate_limiter=app_rate_limiter, request_error_handling=request_error_handling, http_client=client),
        SpectatorAPI(api_key, app_rate_limiter=app_rate_limiter, request_error_handling=request_error_handling, http_client=client),
        StatusAPI(api_key, app_rate_limiter=app_rate_limiter, request_error_handling=request_error_handling, http_client=client),
        LeaguesAPI(api_key, app_rate_limiter=app_rate_limiter, request_error_handling=request_error_handling, http_client=client),
        ThirdPartyCodeAPI(api_key, app_rate_limiter=app_rate_limiter, request_error_handling=request_error_handling, http_client=client)
    }

    return services


class RiotAPI(CompositeDataSource):
    def __init__(self, api_key: str = None, services: Iterable[RiotAPIService] = None, limiting_share: float = 1.0, request_error_handling: Dict = None) -> None:
        if api_key is None:
            api_key = "RIOT_API_KEY"  # Use this env variable.
        if not api_key.startswith("RGAPI"):
            api_key = os.environ.get(api_key, None)

        if services is None:
            services = _default_services(api_key=api_key, limiting_share=limiting_share, request_error_handling=request_error_handling)

        super().__init__(services)

    def set_api_key(self, key: str):
        for sources in self._sources.values():
            for source in sources:
                if isinstance(source, RiotAPIService):
                    source._headers["X-Riot-Token"] = key



from typing import Type, TypeVar, MutableMapping, Any, Iterable, Generator

from datapipelines import DataSource, PipelineContext, Query, NotFoundError, validate_query
from .common import KernelSource, APINotFoundError
from ...data import Platform, Region
from ...dto.staticdata.version import VersionListDto
from ...dto.status import ShardStatusDto
from ..uniquekeys import convert_region_to_platform

T = TypeVar("T")


def _get_default_version(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    pipeline = context[PipelineContext.Keys.PIPELINE]
    versions = pipeline.get(VersionListDto, {"platform": query["platform"]})
    return versions["versions"][0]


def _get_default_locale(query: MutableMapping[str, Any], context: PipelineContext) -> str:
    return query["platform"].default_locale


class StatusAPI(KernelSource):
    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    ##########
    # Status #
    ##########

    _validate_get_status_query = Query. \
        has("platform").as_(Platform)

    @get.register(ShardStatusDto)
    @validate_query(_validate_get_status_query, convert_region_to_platform)
    def get_status(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ShardStatusDto:
        parameters = {"platform": query["platform"].value}
        endpoint = "lol/status/v3/shard-data".format()
        try:
            data = self._get(endpoint=endpoint, parameters=parameters)
        except APINotFoundError as error:
            raise NotFoundError(str(error)) from error

        data["region"] = query["platform"].region.value
        return ShardStatusDto(data)

    _validate_get_many_status_query = Query. \
        has("platforms").as_(Iterable)

    @get_many.register(ShardStatusDto)
    @validate_query(_validate_get_many_status_query, convert_region_to_platform)
    def get_many_status(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Generator[ShardStatusDto, None, None]:
        def generator():
            parameters = {"platform": query["platform"].value}
            for platform in query["platforms"]:
                platform = Platform(platform.upper())
                endpoint = "lol/status/v3/shard-data"
                try:
                    data = self._get(endpoint=endpoint, parameters=parameters)
                except APINotFoundError as error:
                    raise NotFoundError(str(error)) from error

                data["region"] = platform.region.value
                yield ShardStatusDto(data)

        return generator()



import arrow
from typing import Union

from merakicommons.cache import lazy, lazy_property
from merakicommons.container import searchable

from ..data import Region, Platform
from .common import CoreData, CassiopeiaObject, CassiopeiaGhost, CassiopeiaLazyList, CoreDataList, get_latest_version, provide_default_region, ghost_load_on
from ..dto.championmastery import ChampionMasteryDto
from .staticdata.champion import Champion
from .summoner import Summoner
from ..dto import championmastery as dto


##############
# Data Types #
##############


class ChampionMasteryListData(CoreDataList):
    _dto_type = dto.ChampionMasteryListDto
    _renamed = {}


class ChampionMasteryData(CoreData):
    _dto_type = ChampionMasteryDto
    _renamed = {"championLevel": "level", "championPoints": "points", "playerId": "summonerId", "championPointsUntilNextLevel": "pointsUntilNextLevel", "championPointsSinceLastLevel": "pointsSinceLastLevel", "lastPlayTime": "lastPlayed", "tokensEarned": "tokens"}


##############
# Core Types #
##############


class ChampionMasteries(CassiopeiaLazyList):
    _data_types = {ChampionMasteryListData}

    @provide_default_region
    def __init__(self, *, summoner: Summoner, region: Union[Region, str] = None):
        self.__summoner = summoner
        kwargs = {"region": region}
        CassiopeiaObject.__init__(self, **kwargs)

    @classmethod
    @provide_default_region
    def __get_query_from_kwargs__(cls, *, summoner: Union[Summoner, int, str], region: Union[Region, str]) -> dict:
        query = {"region": region}
        if isinstance(summoner, Summoner):
            query["summoner.id"] = summoner.id
        elif isinstance(summoner, str):
            if len(summoner) < 35:
                # It's a summoner name
                query["summoner.id"] = Summoner(name=summoner, region=region).id
            else:
                # It's probably a summoner id
                query["summoner.id"] = summoner
        return query

    @lazy_property
    def region(self) -> Region:
        return Region(self._data[ChampionMasteryListData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @property
    def summoner(self):
        return self.__summoner


@searchable({str: ["champion", "summoner"], int: ["points", "level"], bool: ["chest_granted"], arrow.Arrow: ["last_played"], Champion: ["champion"], Summoner: ["summoner"]})
class ChampionMastery(CassiopeiaGhost):
    _data_types = {ChampionMasteryData}

    @provide_default_region
    def __init__(self, *, summoner: Union[Summoner, int, str] = None, champion: Union[Champion, int, str] = None, region: Union[Region, str] = None, _account_id: str = None):
        kwargs = {"region": region}

        if _account_id is not None:
            summoner = Summoner(account_id=_account_id, region=region)

        if summoner is not None:
            if isinstance(summoner, Summoner):
                self.__class__.summoner.fget._lazy_set(self, summoner)
            elif isinstance(summoner, str):
                if len(summoner) < 35:
                    # It's a summoner name
                    summoner = Summoner(name=summoner, region=region)
                    self.__class__.summoner.fget._lazy_set(self, summoner)
                else:
                    # It's probably a summoner id
                    kwargs["summonerId"] = summoner

        if champion is not None:
            if isinstance(champion, Champion):
                self.__class__.champion.fget._lazy_set(self, champion)
            elif isinstance(champion, str):
                champion = Champion(name=champion, region=self.region, version=get_latest_version(self.region, endpoint="champion"))
                self.__class__.champion.fget._lazy_set(self, champion)
            else:  # int
                kwargs["championId"] = champion

        super().__init__(**kwargs)

    @classmethod
    @provide_default_region
    def __get_query_from_kwargs__(cls, *, summoner: Union[Summoner, int, str], champion: Union[Champion, int, str], region: Union[Region, str]) -> dict:
        query = {"region": region}
        if isinstance(summoner, Summoner):
            query["summoner.id"] = summoner.id
        elif isinstance(summoner, str):
            if len(summoner) < 35:
                # It's a summoner name
                query["summoner.id"] = Summoner(name=summoner, region=region).id
            else:
                # It's probably a summoner id
                query["summoner.id"] = summoner
            

        if isinstance(champion, Champion):
            query["champion.id"] = champion.id
        elif isinstance(champion, str):
            query["champion.id"] = Champion(name=champion, region=region).id
        else:  # int
            query["champion.id"] = champion
        return query

    def __get_query__(self):
        return {"region": self.region, "platform": self.platform.value, "summoner.id": self.summoner.id, "champion.id": self.champion.id}

    def __load__(self, load_group: CoreData = None) -> None:
        from datapipelines import NotFoundError
        try:
            return super().__load__(load_group)
        except NotFoundError:
            from ..transformers.championmastery import ChampionMasteryTransformer
            dto = {
                "championLevel": 0,
                "chestGranted": False,
                "championPoints": 0,
                "championPointsUntilNextLevel": 1800,
                "tokensEarned": 0,
                "championPointsSinceLastLevel": 0,
                "lastPlayTime": None
            }
            data = ChampionMasteryTransformer.champion_mastery_dto_to_data(None, dto)
            self.__load_hook__(load_group, data)

    def __eq__(self, other: "ChampionMastery"):
        if not isinstance(other, ChampionMastery) or self.region != other.region:
            return False
        return self.champion == other.champion and self.summoner == other.summoner

    def __str__(self):
        return "ChampionMastery(summoner={summoner}, champion={champion})".format(summoner=str(self.summoner), champion=str(self.champion))

    __hash__ = CassiopeiaGhost.__hash__

    @property
    def region(self) -> Region:
        return Region(self._data[ChampionMasteryData].region)

    @lazy_property
    def platform(self) -> Platform:
        return self.region.platform

    @CassiopeiaGhost.property(ChampionMasteryData)
    @lazy
    def champion(self) -> Champion:
        """Champion for this entry."""
        return Champion(id=self._data[ChampionMasteryData].championId, region=self.region, version=get_latest_version(self.region, endpoint="champion"))

    @CassiopeiaGhost.property(ChampionMasteryData)
    @lazy
    def summoner(self) -> Summoner:
        """Summoner for this entry."""
        return Summoner(id=self._data[ChampionMasteryData].summonerId, region=self.region)

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on
    def chest_granted(self) -> bool:
        """Is chest granted for this champion or not in current season?"""
        return self._data[ChampionMasteryData].chestGranted

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on
    def level(self) -> int:
        """Champion level for specified player and champion combination."""
        return self._data[ChampionMasteryData].level

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on
    def points(self) -> int:
        """Total number of champion points for this player and champion combination - used to determine champion level."""
        return self._data[ChampionMasteryData].points

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on
    def points_until_next_level(self) -> int:
        """Number of points needed to achieve next level. Zero if player reached maximum champion level for this champion."""
        return self._data[ChampionMasteryData].pointsUntilNextLevel

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on
    def tokens(self) -> int:
        """Number of tokens earned toward next mastery level."""
        return self._data[ChampionMasteryData].tokens

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on
    def points_since_last_level(self) -> int:
        """Number of points earned since current level has been achieved. Zero if player reached maximum champion level for this champion."""
        return self._data[ChampionMasteryData].pointsSinceLastLevel

    @CassiopeiaGhost.property(ChampionMasteryData)
    @ghost_load_on
    @lazy
    def last_played(self) -> arrow.Arrow:
        """Last time this champion was played by this player."""
        return arrow.get(self._data[ChampionMasteryData].lastPlayed / 1000)



from typing import Type, TypeVar, MutableMapping, Any, Iterable, Union
import arrow
import copy

from datapipelines import DataSource, PipelineContext, Query, validate_query

from ..data import Platform, Queue, Tier, Division
from ..core import Champion, Rune, Item, Map, SummonerSpell, Realms, ProfileIcon, LanguageStrings, Summoner, ChampionMastery, Match, CurrentMatch, ShardStatus, ChallengerLeague, GrandmasterLeague, MasterLeague, League, MatchHistory, Items, Champions, Maps, ProfileIcons, Locales, Runes, SummonerSpells, Versions, ChampionMasteries, LeagueEntries, FeaturedMatches, VerificationString
from ..core.match import Timeline, MatchListData
from ..core.championmastery import ChampionMasteryListData
from ..core.league import LeagueEntry, LeagueEntriesData, LeagueEntryData, LeagueSummonerEntries, LeagueSummonerEntriesData
from ..core.spectator import FeaturedGamesData
from ..core.staticdata.item import ItemListData
from ..core.staticdata.champion import ChampionListData, ChampionData
from ..core.staticdata.map import MapListData
from ..core.staticdata.profileicon import ProfileIconListData
from ..core.staticdata.language import LanguagesData
from ..core.staticdata.rune import RuneListData
from ..core.staticdata.summonerspell import SummonerSpellListData
from ..core.staticdata.version import VersionListData
from .riotapi.common import _get_latest_version, _get_default_locale
from .uniquekeys import convert_region_to_platform

T = TypeVar("T")


class UnloadedGhostStore(DataSource):
    def __init__(self):
        super().__init__()

    @DataSource.dispatch
    def get(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> T:
        pass

    @DataSource.dispatch
    def get_many(self, type: Type[T], query: MutableMapping[str, Any], context: PipelineContext = None) -> Iterable[T]:
        pass

    _validate_get_versions_query = Query. \
        has("platform").as_(Platform)

    _validate_get_realms_query = Query. \
        has("platform").as_(Platform)

    _validate_get_champion_query = Query. \
        has("platform").as_(Platform).also. \
        has("id").as_(int).or_("name").as_(str).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").also. \
        can_have("includedData")

    _validate_get_champions_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    _validate_get_rune_query = Query. \
        has("id").as_(int).or_("name").as_(str).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    _validate_get_runes_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    _validate_get_item_query = Query. \
        has("id").as_(int).or_("name").as_(str).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    _validate_get_items_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    _validate_get_map_query = Query. \
        has("id").as_(int).or_("name").as_(str).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    _validate_get_maps_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    _validate_get_summoner_spell_query = Query. \
        has("id").as_(int).or_("name").as_(str).also. \
        has("platform").as_(Platform).also. \
        can_have("version").with_default(_get_latest_version, supplies_type=str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    _validate_get_summoner_spells_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str).also. \
        can_have("includedData").with_default({"all"})

    _validate_get_languages_query = Query. \
        has("platform").as_(Platform)

    _validate_get_language_strings_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    _validate_get_profile_icon_query = Query. \
        has("platform").as_(Platform).also. \
        has("id").as_(int).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    _validate_get_profile_icons_query = Query. \
        has("platform").as_(Platform).also. \
        can_have("version").as_(str).also. \
        can_have("locale").with_default(_get_default_locale, supplies_type=str)

    _validate_get_champion_mastery_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(str).also. \
        has("champion.id").as_(int)

    _validate_get_champion_masteries_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(str)

    _validate_get_paginated_queues_query = Query. \
        has("platform").as_(Platform)

    _validate_get_league_entries_query = Query. \
        has("tier").as_(Tier).also. \
        has("division").as_(Division).also. \
        has("queue").as_(Queue).also. \
        has("platform").as_(Platform)

    _validate_get_league_summoner_entries_query = Query. \
        has("summoner.id").as_(str).also. \
        has("platform").as_(Platform)

    _validate_get_league_query = Query. \
        has("id").as_(str).also. \
        has("platform").as_(Platform)

    _validate_get_challenger_league_query = Query. \
        has("queue").as_(Queue).also. \
        has("platform").as_(Platform)

    _validate_get_grandmaster_league_query = Query. \
        has("queue").as_(Queue).also. \
        has("platform").as_(Platform)

    _validate_get_master_league_query = Query. \
        has("queue").as_(Queue).also. \
        has("platform").as_(Platform)

    _validate_get_league_entries_list_query = Query. \
        has("queue").as_(Queue).also. \
        has("tier").as_(Tier).also. \
        has("division").as_(Division).also. \
        has("platform").as_(Platform)

    _validate_get_current_match_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(str)

    _validate_get_featured_matches_query = Query. \
        has("platform").as_(Platform)

    _validate_get_match_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform)

    _validate_get_match_history_query = Query. \
        has("accountId").as_(str).also. \
        has("platform").as_(Platform).also. \
        can_have("beginTime").as_(int).also. \
        can_have("endTime").as_(int).also. \
        can_have("beginIndex").as_(int).also. \
        can_have("endIndex").as_(int).also. \
        can_have("seasons").as_(Iterable).also. \
        can_have("champion.ids").as_(Iterable).also. \
        can_have("queues").as_(Iterable)

    _validate_get_timeline_query = Query. \
        has("id").as_(int).also. \
        has("platform").as_(Platform)

    _validate_get_shard_status_query = Query. \
        has("platform").as_(Platform)

    _validate_get_summoner_query = Query. \
        has("id").as_(str). \
        or_("accountId").as_(str). \
        or_("puuid").as_(str). \
        or_("name").as_(str).also. \
        has("platform").as_(Platform)

    _validate_get_verification_string_query = Query. \
        has("platform").as_(Platform).also. \
        has("summoner.id").as_(str)

    @get.register(Champion)
    @validate_query(_validate_get_champion_query, convert_region_to_platform)
    def get_champion(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Champion:
        query["region"] = query.pop("platform").region
        return Champion._construct_normally(**query)

    @get.register(Rune)
    @validate_query(_validate_get_rune_query, convert_region_to_platform)
    def get_rune(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Rune:
        query["region"] = query.pop("platform").region
        query["includedData"] = query.pop("includedData")
        return Rune._construct_normally(**query)

    @get.register(Item)
    @validate_query(_validate_get_item_query, convert_region_to_platform)
    def get_item(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Item:
        query["region"] = query.pop("platform").region
        query["includedData"] = query.pop("includedData")
        return Item._construct_normally(**query)

    @get.register(Map)
    @validate_query(_validate_get_map_query, convert_region_to_platform)
    def get_map(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Map:
        query["region"] = query.pop("platform").region
        return Map._construct_normally(**query)

    @get.register(SummonerSpell)
    @validate_query(_validate_get_summoner_spell_query, convert_region_to_platform)
    def get_summoner_spell(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> SummonerSpell:
        query["region"] = query.pop("platform").region
        query["includedData"] = query.pop("includedData")
        return SummonerSpell._construct_normally(**query)

    @get.register(Realms)
    @validate_query(_validate_get_realms_query, convert_region_to_platform)
    def get_realms(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Realms:
        query["region"] = query.pop("platform").region
        return Realms._construct_normally(**query)

    @get.register(ProfileIcon)
    @validate_query(_validate_get_profile_icon_query, convert_region_to_platform)
    def get_profile_icons(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ProfileIcon:
        query["region"] = query.pop("platform").region
        return ProfileIcon._construct_normally(**query)

    @get.register(LanguageStrings)
    @validate_query(_validate_get_language_strings_query, convert_region_to_platform)
    def get_language_strings(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LanguageStrings:
        query["region"] = query.pop("platform").region
        return LanguageStrings._construct_normally(**query)

    @get.register(Summoner)
    @validate_query(_validate_get_summoner_query, convert_region_to_platform)
    def get_summoner(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Summoner:
        kwargs = copy.deepcopy(query)
        kwargs["region"] = kwargs.pop("platform").region
        if "accountId" in kwargs:
            kwargs["account_id"] = kwargs.pop("accountId")
        return Summoner._construct_normally(**kwargs)

    @get.register(ChampionMastery)
    @validate_query(_validate_get_champion_mastery_query, convert_region_to_platform)
    def get_champion_mastery(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionMastery:
        query["region"] = query.pop("platform").region
        if "summoner.id" in query:
            query["summoner"] = query.pop("summoner.id")
        if "summoner.name" in query:
            query["summoner"] = query.pop("summoner.name")
        if "summoner.accountId" in query:
            query["_account_id"] = query.pop("summoner.accountId")
        if "champion.id" in query:
            query["champion"] = query.pop("champion.id")
        if "champion.id" in query:
            query["champion"] = query.pop("champion.name")
        return ChampionMastery._construct_normally(**query)

    @get.register(Match)
    @validate_query(_validate_get_match_query, convert_region_to_platform)
    def get_match(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Match:
        query["region"] = query.pop("platform").region
        return Match._construct_normally(**query)

    @get.register(Timeline)
    @validate_query(_validate_get_timeline_query, convert_region_to_platform)
    def get_match(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Timeline:
        query["region"] = query.pop("platform").region
        return Timeline._construct_normally(**query)

    @get.register(CurrentMatch)
    @validate_query(_validate_get_current_match_query, convert_region_to_platform)
    def get_current_match(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> CurrentMatch:
        query["region"] = query.pop("platform").region
        query["summoner"] = query.pop("summoner.id")
        return CurrentMatch._construct_normally(**query)

    @get.register(ShardStatus)
    @validate_query(_validate_get_shard_status_query, convert_region_to_platform)
    def get_shard_status(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ShardStatus:
        query["region"] = query.pop("platform").region
        return ShardStatus._construct_normally(**query)

    @get.register(ChallengerLeague)
    @validate_query(_validate_get_challenger_league_query, convert_region_to_platform)
    def get_challenger_league(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChallengerLeague:
        UnloadedGhostStore._validate_get_challenger_league_query(query)
        query["region"] = query.pop("platform").region
        return ChallengerLeague._construct_normally(**query)

    @get.register(GrandmasterLeague)
    @validate_query(_validate_get_grandmaster_league_query, convert_region_to_platform)
    def get_grandmaster_league(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> GrandmasterLeague:
        UnloadedGhostStore._validate_get_grandmaster_league_query(query)
        query["region"] = query.pop("platform").region
        return GrandmasterLeague._construct_normally(**query)

    @get.register(MasterLeague)
    @validate_query(_validate_get_master_league_query, convert_region_to_platform)
    def get_master_league(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MasterLeague:
        query["region"] = query.pop("platform").region
        return MasterLeague._construct_normally(**query)

    @get.register(League)
    @validate_query(_validate_get_league_query, convert_region_to_platform)
    def get_league(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> League:
        query["region"] = query.pop("platform").region
        query["id"] = query.pop("id")
        return League._construct_normally(**query)

    @get.register(LeagueEntries)
    @validate_query(_validate_get_league_entries_list_query, convert_region_to_platform)
    def get_league_entries_list(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LeagueEntries:
        def generate_entries(original_query):
            page = 1
            while True:
                new_query = copy.deepcopy(original_query)
                new_query["page"] = page
                data = context[context.Keys.PIPELINE].get(LeagueEntriesData, query=new_query)
                n_new_results = len(data)
                for entrydata in data:
                    entry = LeagueEntry.from_data(data=entrydata, loaded_groups={LeagueEntryData})
                    yield entry
                if page == 1:
                    results_per_page = n_new_results
                if n_new_results != results_per_page:
                    break
                page += 1

        original_query = copy.deepcopy(query)
        return LeagueEntries.from_generator(generator=generate_entries(original_query), region=query["region"], queue=query["queue"], tier=query["tier"], division=query["division"])

    @get.register(LeagueSummonerEntries)
    @validate_query(_validate_get_league_summoner_entries_query, convert_region_to_platform)
    def get_league_summoner_entries(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> LeagueEntries:
        def league_summoner_entries_generator(query):
            data = context[context.Keys.PIPELINE].get(LeagueSummonerEntriesData, query)
            for entry in data:
                entry = LeagueEntry.from_data(entry)
                yield entry
        kwargs = {
            "summoner": Summoner(id=query["summoner.id"], region=query["region"])
        }
        return LeagueSummonerEntries.from_generator(generator=league_summoner_entries_generator(query), **kwargs)

    @get.register(VerificationString)
    @validate_query(_validate_get_verification_string_query, convert_region_to_platform)
    def get_verification_string(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> VerificationString:
        query["region"] = query.pop("platform").region
        query["summoner"] = Summoner(id=query.pop("summoner.id"), region=query["region"])
        return VerificationString._construct_normally(**query)

    @get.register(MatchHistory)
    @validate_query(_validate_get_match_history_query, convert_region_to_platform)
    def get_match_history(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> MatchHistory:
        original_query = copy.deepcopy(query)
        region = query["region"]
        account_id = query["accountId"]
        begin_index = query.get("beginIndex", 0)
        end_index = query.get("endIndex", None)
        first_requested_dt = query.get("beginTime", 0)  # (begin_time) Defaults to start of summoner's match history
        most_recent_requested_dt = query.get("endTime", None)  # (end_time) Defaults to now; end_time > begin_time
        queues = query.get("queues", {})
        seasons = query.get("seasons", {})
        champion_ids = query.get("champion.ids", {})

        max_number_of_requested_matches = float("inf") if end_index is None else end_index - begin_index

        # Convert times to datetimes
        first_requested_dt = arrow.get(first_requested_dt / 1000)
        if most_recent_requested_dt is not None:
            most_recent_requested_dt = arrow.get(most_recent_requested_dt / 1000)
            assert most_recent_requested_dt > first_requested_dt

        # Create the generator that will populate the match history object.
        def generate_matchlists(begin_index: int, max_number_of_requested_matches: int = None, first_requested_dt: arrow.Arrow = None, most_recent_requested_dt: Union[arrow.Arrow, None] = None):
            _begin_index = begin_index
            _current_first_requested_dt = first_requested_dt

            if isinstance(max_number_of_requested_matches, int):
                max_number_of_requested_matches = float(max_number_of_requested_matches)

            pulled_matches = 0
            while pulled_matches < max_number_of_requested_matches:
                new_query = {
                    "region": region,
                    "accountId": account_id,
                    "queues": queues,
                    "seasons": seasons,
                    "champion.ids": champion_ids,
                    "beginIndex": _begin_index,
                    "beginTime": _current_first_requested_dt.timestamp * 1000,
                    "maxNumberOfMatches": max_number_of_requested_matches
                }
                if "endTime" in original_query:
                    new_query["endTime"] = original_query["endTime"]

                data = context[context.Keys.PIPELINE].get(MatchListData, query=new_query)
                matchrefdata = None
                for matchrefdata in data:
                    pulled_matches += 1
                    if pulled_matches > 0 and (most_recent_requested_dt is None or matchrefdata.creation < most_recent_requested_dt) and matchrefdata.creation > first_requested_dt:
                        match = Match.from_match_reference(matchrefdata)
                        yield  match
                    if pulled_matches >= max_number_of_requested_matches:
                        break
                    if first_requested_dt is not None and matchrefdata.creation <= first_requested_dt:
                        break

                matches_pulled_this_iteration = len(data)
                if hasattr(data, "endIndex"):
                    expected_pulled_matches_this_iteration = data.endIndex - data.beginIndex
                else:
                    expected_pulled_matches_this_iteration = None
                if expected_pulled_matches_this_iteration is not None and matches_pulled_this_iteration < expected_pulled_matches_this_iteration:
                    # Stop because the API returned less data than we asked for, and so there isn't any more left
                    break

                if most_recent_requested_dt is not None and matchrefdata is not None and matchrefdata.creation >= most_recent_requested_dt:
                    # Stop because we have gotten all the matches in the requested time range
                    break

                # Update the "start" values for the next iteration of calls
                if hasattr(data, "endIndex"):
                    _begin_index = data.endIndex
                # else: Don't update _begin_index
                if hasattr(data, "beginTime"):
                    _current_first_requested_dt = arrow.get(data.beginTime / 1000)
                    _earliest_dt_in_current_match_history = _current_first_requested_dt
                else:
                    _earliest_dt_in_current_match_history = matchrefdata.creation
                # else: Don't update _current_first_requested_dt
                max_number_of_requested_matches -= len(data)

                if first_requested_dt is not None and _earliest_dt_in_current_match_history <= first_requested_dt:
                    # Stop because we have gotten all the matches in the requested time range
                    break

        generator = generate_matchlists(begin_index, max_number_of_requested_matches, first_requested_dt, most_recent_requested_dt)

        summoner = Summoner(account_id=account_id, region=region)
        generator = MatchHistory.from_generator(generator=generator, summoner=summoner,
                                                begin_index=begin_index, end_index=end_index,
                                                begin_time=first_requested_dt,
                                                end_time=most_recent_requested_dt,
                                                queues=queues, seasons=seasons, champions=champion_ids)
        return generator

    @get.register(Items)
    @validate_query(_validate_get_items_query, convert_region_to_platform)
    def get_items(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Items:
        def items_generator(query):
            data = context[context.Keys.PIPELINE].get(ItemListData, query)
            for itemdata in data:
                item = Item.from_data(itemdata)
                yield item
        kwargs = {
            "region": query["region"],
            "version": query["version"],
            "locale": query["locale"],
            "included_data": query["includedData"],
        }
        return Items.from_generator(generator=items_generator(query), **kwargs)

    @get.register(Champions)
    @validate_query(_validate_get_champions_query, convert_region_to_platform)
    def get_champions(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Champions:
        def champions_generator(query):
            data = context[context.Keys.PIPELINE].get(ChampionListData, query)
            for champion_data in data:
                champion = Champion.from_data(champion_data, loaded_groups={ChampionData})
                yield champion
        kwargs = {
            "region": query["region"],
            "version": query["version"],
            "locale": query["locale"],
            "included_data": query["includedData"],
        }
        return Champions.from_generator(generator=champions_generator(query), **kwargs)

    @get.register(Maps)
    @validate_query(_validate_get_maps_query, convert_region_to_platform)
    def get_maps(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Maps:
        def maps_generator(query):
            data = context[context.Keys.PIPELINE].get(MapListData, query)
            for map_data in data:
                map = Map.from_data(map_data)
                yield map
        kwargs = {
            "region": query["region"],
            "version": query["version"],
            "locale": query["locale"]
        }
        return Maps.from_generator(generator=maps_generator(query), **kwargs)

    @get.register(ProfileIcons)
    @validate_query(_validate_get_profile_icons_query, convert_region_to_platform)
    def get_profile_icons(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ProfileIcons:
        def profile_icons_generator(query):
            data = context[context.Keys.PIPELINE].get(ProfileIconListData, query)
            for profile_icon_data in data:
                profile_icon = ProfileIcon.from_data(profile_icon_data)
                yield profile_icon
        kwargs = {
            "region": query["region"],
            "version": query["version"],
            "locale": query["locale"]
        }
        return ProfileIcons.from_generator(generator=profile_icons_generator(query), **kwargs)

    @get.register(Locales)
    @validate_query(_validate_get_languages_query, convert_region_to_platform)
    def get_locales(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Locales:
        def locales_generator(query):
            data = context[context.Keys.PIPELINE].get(LanguagesData, query)
            for locale in data:
                yield locale
        kwargs = {
            "region": query["region"]
        }
        return Locales.from_generator(generator=locales_generator(query), **kwargs)

    @get.register(Runes)
    @validate_query(_validate_get_runes_query, convert_region_to_platform)
    def get_runes(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Runes:
        def runes_generator(query):
            data = context[context.Keys.PIPELINE].get(RuneListData, query)
            for runedata in data:
                rune = Rune.from_data(runedata)
                yield rune
        kwargs = {
            "region": query["region"],
            "version": query["version"],
            "locale": query["locale"],
            "included_data": query["includedData"],
        }
        return Runes.from_generator(generator=runes_generator(query), **kwargs)

    @get.register(SummonerSpells)
    @validate_query(_validate_get_summoner_spells_query, convert_region_to_platform)
    def get_summoner_spells(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> SummonerSpells:
        def summoner_spells_generator(query):
            data = context[context.Keys.PIPELINE].get(SummonerSpellListData, query)
            for summoner_spelldata in data:
                summoner_spell = SummonerSpell.from_data(summoner_spelldata)
                yield summoner_spell
        kwargs = {
            "region": query["region"],
            "version": query["version"],
            "locale": query["locale"],
            "included_data": query["includedData"],
        }
        return SummonerSpells.from_generator(generator=summoner_spells_generator(query), **kwargs)

    @get.register(Versions)
    @validate_query(_validate_get_versions_query, convert_region_to_platform)
    def get_versions(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> Versions:
        def versions_generator(query):
            data = context[context.Keys.PIPELINE].get(VersionListData, query)
            for version in data:
                yield version
        kwargs = {
            "region": query["region"]
        }
        return Versions.from_generator(generator=versions_generator(query), **kwargs)

    @get.register(ChampionMasteries)
    @validate_query(_validate_get_champion_masteries_query, convert_region_to_platform)
    def get_versions(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> ChampionMasteries:
        def champion_masteries_generator(query):
            from ..transformers.championmastery import ChampionMasteryTransformer
            all_champion_ids = [champion.id for champion in Champions(region=query["region"])]
            data = context[context.Keys.PIPELINE].get(ChampionMasteryListData, query)
            for champion_mastery_data in data:
                champion_mastery = ChampionMastery.from_data(champion_mastery_data)
                all_champion_ids.remove(champion_mastery.champion.id)
                yield champion_mastery
            for unfound_id in all_champion_ids:
                dto = {
                    "championId": unfound_id,
                    "playerId": query["summoner.id"],
                    "championLevel": 0,
                    "chestGranted": False,
                    "championPoints": 0,
                    "championPointsUntilNextLevel": 1800,
                    "championPointsSinceLastLevel": 0,
                    "lastPlayTime": None,
                    "region": query["region"]
                }
                champion_mastery_data = ChampionMasteryTransformer.champion_mastery_dto_to_data(None, dto)
                champion_mastery = ChampionMastery.from_data(champion_mastery_data)
                yield champion_mastery
        kwargs = {
            "region": query["region"],
            "summoner": Summoner(id=query["summoner.id"], region=query["region"])
        }
        return ChampionMasteries.from_generator(generator=champion_masteries_generator(query), **kwargs)



    @get.register(FeaturedMatches)
    @validate_query(_validate_get_featured_matches_query, convert_region_to_platform)
    def get_featured_matches(self, query: MutableMapping[str, Any], context: PipelineContext = None) -> FeaturedMatches:
        def featured_matches_generator(query):
            data = context[context.Keys.PIPELINE].get(FeaturedGamesData, query)
            for match_data in data:
                match = CurrentMatch.from_data(match_data, summoner=None)
                yield match
        kwargs = {
            "region": query["region"]
        }
        return FeaturedMatches.from_generator(generator=featured_matches_generator(query), **kwargs)



# Initialize the settings singleton
from ._configuration import get_default_config, Settings, CassiopeiaConfiguration as _CassiopeiaConfiguration
configuration = _CassiopeiaConfiguration()

from .cassiopeia import get_realms, get_challenger_league, get_champion_masteries, get_champion, get_champion_mastery, get_champions, get_current_match, get_featured_matches, get_items, get_language_strings, get_locales, get_league_entries, get_maps, get_master_league, get_grandmaster_league, get_match, get_match_history, get_profile_icons, get_runes, get_status, get_summoner, get_summoner_spells, get_version, get_versions, get_champion_rotations, get_paginated_league_entries, get_verification_string
from .cassiopeia import apply_settings, set_riot_api_key, set_default_region, print_calls, _get_pipeline
from .core import Champion, Champions, Rune, Runes, Item, Items, SummonerSpell, SummonerSpells, ProfileIcon, ProfileIcons, Versions, Maps, Summoner, ChampionMastery, ChampionMasteries, Match, FeaturedMatches, ShardStatus, ChallengerLeague, GrandmasterLeague, MasterLeague, Map, Realms, LanguageStrings, Locales, LeagueEntries, League, Patch, VerificationString, MatchHistory, ChampionRotation, LeagueSummonerEntries, CurrentMatch
from .data import Queue, Region, Platform, Resource, Side, GameMode, MasteryTree, Tier, Division, Season, GameType, Lane, Role, Rank, Key, SummonersRiftArea, Tower, Position

# apply_settings(configuration.settings)
# This is inicialized on django_cassiopeia __init__
